import numpy as np
from utils import *
from graphics.texture import *
from utils import *
from .texture import *

class MeshMaterial:
    """
    A class that contains a Mesh Material for applying into objects.
    It is essentially a wrapper of threejs' 'MeshPhysicalMaterial' class.
    See https://threejs.org/docs/#api/en/materials/MeshPhysicalMaterial for
    more details.

    Parameters
    ----------
    opacity : float between 0 and 1
        Mirrors 'Material.opacity' in threejs.
        Float in the range of 0.0 - 1.0 indicating how transparent the material is.
        A value of 0.0 indicates fully transparent, 1.0 is fully opaque.
        The attribute 'transparent' is set to true automatically if opacity<1.
        (default: 1).

    shadow_side : string
        Mirrors 'Material.shadowSide' in threejs.
        Defines which side of faces cast shadows. When set, can be
        ''FrontSide', 'BackSide', or 'DoubleSide' or 'null'.
        (default: 'null').

    side : string
        Mirrors 'Material.side' in threejs.
        Defines which side of faces will be rendered - front, back or both.
        Default is 'FrontSide'. Other options are 'BackSide' and 'DoubleSide'.
        (default: 'FrontSide').

    alpha_map : string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.alphaMap' in threejs.
        The alpha map is a grayscale texture that controls the opacity across the surface
        (black: fully transparent; white: fully opaque).
        If it is a string, a default texture is created from it.
        If it is the empty string, it is ignored.
        (default: empty string).

    ao_map : string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.aoMap' in threejs.
        The red channel of this texture is used as the ambient occlusion map.
        The aoMap requires a second set of UVs.
        If it is a string, a default texture is created from it.
        If it is the empty string, it is ignored.
        (default: empty string).

    ao_map_intensity : float between 0 and 1
        Mirrors 'MeshStandardMaterial.aoMapIntensity' in threejs.
        Intensity of the ambient occlusion effect. Default is 1. Zero is no occlusion effect.
        (default: 1).

    ao_map_intensity : float between 0 and 1
        Mirrors 'MeshStandardMaterial.aoMapIntensity' in threejs.
        Intensity of the ambient occlusion effect. Default is 1. Zero is no occlusion effect.
        (default: 1).

    bump_map : string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.bumpMap' in threejs.
        The texture to create a bump map. The black and white values map to the perceived depth in relation to the lights.
        Bump doesn't actually affect the geometry of the object, only the lighting. If a normal map is defined this will
        be ignored.
        If it is a string, a default texture is created from it.
        If it is the empty string, it is ignored.
        (default: empty string).

    bump_scale : float between 0 and 1
        Mirrors 'MeshStandardMaterial.bumpScale' in threejs.
        How much the bump map affects the material.
        (default: 1).

    color : HTML-compatible color string
        Mirrors 'MeshStandardMaterial.color' in threejs.
        Color of the material.
        (default: "white").

    emissive : HTML-compatible color string
        Mirrors 'MeshStandardMaterial.emissive' in threejs.
        Emissive (light) color of the material, essentially a solid color unaffected by other lighting.
        (default: "black").

    emissive_map : string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.emissiveMap' in threejs.
        Set emisssive (glow) map. Default is null. The emissive map color is modulated by the emissive color and the
        emissive intensity. If you have an emissive map, be sure to set the emissive color to something other
        than black.
        (default: empty string).

    emissive_intensity : positive float
        Mirrors 'MeshStandardMaterial.emissiveIntensity' in threejs.
        Intensity of the emissive light. Modulates the emissive color.
        (default: 1).

    env_map : string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.envMap' in threejs.
        The environment map. To ensure a physically correct rendering, you should only add environment maps which were
        preprocessed by 'PMREMGenerator'.
        (default: empty string).

    env_map_intensity : positive float
        Mirrors 'MeshStandardMaterial.envMapIntensity' in threejs.
        Scales the effect of the environment map by multiplying its color.
        (default: 1).

    flat_shading : boolean
        Mirrors 'MeshStandardMaterial.flatShading' in threejs.
        Define whether the material is rendered with flat shading.
        (default: False).

    light_map : string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.lightMap' in threejs.
        The light map. The lightMap requires a second set of UVs.
        (default: empty string).

    light_map_intensity : positive float
        Mirrors 'MeshStandardMaterial.lightMapIntensity' in threejs.
        Intensity of the baked light.
        (default: 1).

    texture_map: string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.map' in threejs.
        The color map. The texture map color is modulated by the diffuse .color.
        (Obs: the name of this parameter in 'Python' is not 'map' because it is
        a reserved word).
        (default: empty string).

    metalness : float between 0 and 1
        Mirrors 'MeshStandardMaterial.metalness' in threejs.
        How much the material is like a metal. Non-metallic materials such as wood or stone use 0.0, metallic use 1.0,
        with nothing (usually) in between. A value between 0.0 and 1.0 could be used for a rusty metal
        look. If metalnessMap is also provided, both values are multiplied.
        (default: 0).

    metalness_map: string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.metalnessMap' in threejs.
        The blue channel of this texture is used to alter the metalness of the material.
        (default: empty string).

    normal_map: string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.normalMap' in threejs.
        The texture to create a normal map. The RGB values affect the surface normal for each pixel fragment and change
        the way the color is lit. Normal maps do not change the actual shape of the surface, only the lighting. In case
        the material has a normal map authored using the left handed convention, the y component of normalScale should
        be negated to compensate for the different handedness.
        (default: empty string).

    normal_scale: 2D vector
        Mirrors 'MeshStandardMaterial.normalScale' in threejs.
        How much the normal map affects the material. Typical ranges are 0-1.
        (default: [1,1]).

    refraction_ratio : float between 0 and 1
        Mirrors 'MeshStandardMaterial.refractionRatio' in threejs.
        The index of refraction (IOR) of air (approximately 1) divided by the index of refraction of the material. It is
        used with environment mapping modes 'CubeRefractionMapping' and 'EquirectangularRefractionMapping'.
        The refraction ratio should not exceed 1.
        (default: 0.98).

    roughness : float between 0 and 1
        Mirrors 'MeshStandardMaterial.roughness' in threejs.
        How rough the material appears. 0.0 means a smooth mirror reflection, 1.0 means fully diffuse.
        If roughnessMap is also provided, both values are multiplied.
        (default: 1).

    normal_map: string containing an url or a 'Texture' object
        Mirrors 'MeshStandardMaterial.roughnessMap' in threejs.
        The green channel of this texture is used to alter the roughness of the material.
        (default: empty string).

    clearcoat : float between 0 and 1
        Mirrors 'MeshPhysicalMaterial.clearcoat' in threejs.
        Represents the intensity of the clear coat layer, from 0.0 to 1.0. Use clear coat related properties to enable
        multilayer materials that have a thin translucent layer over the base layer.
        (default: 0).

    clearcoat_map: string containing an url or a 'Texture' object
        Mirrors 'MeshPhysicalMaterial.clearcoatMap' in threejs.
        The red channel of this texture is multiplied against .clearcoat, for per-pixel control over a coating's
        intensity.
        (default: empty string).

    clearcoat_normal_map: string containing an url or a 'Texture' object
        Mirrors 'MeshPhysicalMaterial.clearcoatNormalMap' in threejs.
        Can be used to enable independent normals for the clear coat layer.
        (default: empty string).

    clearcoat_normal_scale: 2D vector
        Mirrors 'MeshPhysicalMaterial.clearcoatNormalScale' in threejs.
        How much .clearcoatNormalMap affects the clear coat layer, from [0,0] to [1,1].
        (default: [1,1]).

    clearcoat_roughness : float between 0 and 1
        Mirrors 'MeshPhysicalMaterial.clearcoat_roughness' in threejs.
        Roughness of the clear coat layer, from 0.0 to 1.0.
        (default: 0).

    clearcoat_roughness_map: string containing an url or a 'Texture' object
        Mirrors 'MeshPhysicalMaterial.clearcoatRoughnessMap' in threejs.
        The green channel of this texture is multiplied against .clearcoatRoughness, for per-pixel control over a
        coating's roughness.
        (default: empty string).

    ior : float between 1 and 2.333
        Mirrors 'MeshPhysicalMaterial.ior' in threejs.
        Index-of-refraction for non-metallic materials, from 1.0 to 2.333.
        (default: 1.5).

    reflectivity : float between 0 and 1
        Mirrors 'MeshPhysicalMaterial.reflectivity' in threejs.
        Degree of reflectivity, from 0.0 to 1.0. Default is 0.5, which corresponds to an index-of-refraction of 1.5.
        This models the reflectivity of non-metallic materials. It has no effect when metalness is 1.0
        (default: 0.5).

    sheen : float between 0 and 1
        Mirrors 'MeshPhysicalMaterial.sheen' in threejs.
        The intensity of the sheen layer, from 0.0 to 1.0.
        (default: 0).

    sheen_roughness : float between 0 and 1
        Mirrors 'MeshPhysicalMaterial.sheenRoughness' in threejs.
        Roughness of the sheen layer, from 0.0 to 1.0.
        (default: 1).

    sheen_roughness_map: string containing an url or a 'Texture' object
        Mirrors 'MeshPhysicalMaterial.sheenRoughnessMap' in threejs.
        The alpha channel of this texture is multiplied against .sheenRoughness, for per-pixel control over
        sheen roughness.
        (default: empty string).

    sheen_color : HTML-compatible color string
        Mirrors 'MeshPhysicalMaterial.sheenColor' in threejs.
        The sheen tint.
        (default: "white").

    sheen_color_map: string containing an url or a 'Texture' object
        Mirrors 'MeshPhysicalMaterial.sheenColorMap' in threejs.
        The RGB channels of this texture are multiplied against .sheenColor, for per-pixel control over sheen tint.
        (default: empty string).

    specular_intensity : float between 0 and 1
        Mirrors 'MeshPhysicalMaterial.specularIntensity' in threejs.
        A float that scales the amount of specular reflection for non-metals only.
        When set to zero, the model is effectively Lambertian. From 0.0 to 1.0.
        (default: 0).

    specular_intensity_map: string containing an url or a 'Texture' object
        Mirrors 'MeshPhysicalMaterial.specularIntensityMap' in threejs.
        The alpha channel of this texture is multiplied against .specularIntensity, for per-pixel control over
        specular intensity
        (default: empty string).

    specular_color : HTML-compatible color string
        Mirrors 'MeshPhysicalMaterial.specularColor' in threejs.
        A Color that tints the specular reflection at normal incidence for non-metals only.
        (default: "white").

    specular_color_map: string containing an url or a 'Texture' object
        Mirrors 'MeshPhysicalMaterial.specularColorMap' in threejs.
        The RGB channels of this texture are multiplied against .specularColor, for per-pixel control over specular
        color.
        (default: empty string).

    transmission : float between 0 and 1
        Mirrors 'MeshPhysicalMaterial.transmission' in threejs.
        Degree of transmission (or optical transparency), from 0.0 to 1.0.
        Thin, transparent or semitransparent, plastic or glass materials remain largely reflective even if they are
        fully transmissive. The transmission property can be used to model these materials.
        When transmission is non-zero, opacity should be set to 1.
        (default: 0).

    transmission_map: string containing an url or a 'Texture' object
        Mirrors 'MeshPhysicalMaterial.transmissionMap' in threejs.
        The red channel of this texture is multiplied against .transmission, for per-pixel control over optical
        transparency.
        (default: empty string).
    """

    #######################################
    # Attributes
    #######################################

    @property
    def opacity(self) -> float:
        """The object opacity."""
        return self._opacity

    @property
    def shadow_side(self) -> str:
        """Which side of faces casts shadows."""
        return self._shadow_side

    @property
    def side(self) -> str:
        """Which side of faces will be rendered."""
        return self._side

    @property
    def transparent(self) -> bool:
        """If the object is transparent."""
        return self._transparent

    @property
    def alpha_map(self):
        """The alpha map texture"""
        return self._alpha_map

    @property
    def ao_map(self):
        """The texture for the red channel for the ambient occlusion (ao) map."""
        return self._ao_map

    @property
    def ao_map_intensity(self):
        """The ambient occlusion intensity."""
        return self._ao_map_intensity

    @property
    def bump_map(self):
        """The bump texture map."""
        return self._bump_map

    @property
    def bump_map_scale(self):
        """Bump map scale."""
        return self._bump_map_scale

    @property
    def color(self):
        """The object color."""
        return self._color

    @property
    def emissive(self):
        """The emissivity of the material."""
        return self._emissive

    @property
    def emissive_map(self):
        """The emissivity texture map."""
        return self._emissive_map

    @property
    def emissive_intensity(self):
        """The emissivity intensity."""
        return self._emissive_intensity

    @property
    def env_map(self):
        """Environmental texture map."""
        return self._env_map

    @property
    def env_map_intensity(self):
        """The intensivity of the environmental map."""
        return self._env_map_intensity

    @property
    def flat_shading(self):
        """If the object is flat shaded."""
        return self._flat_shading

    @property
    def light_map(self):
        """The texture map of the light."""
        return self._light_map

    @property
    def light_map_itensity(self):
        """The light map intensity."""
        return self._light_map_itensity

    @property
    def map(self):
        """The main texture map."""
        return self._map

    @property
    def metalness(self):
        """The metalness of the object."""
        return self._metalness

    @property
    def metalness_map(self):
        """The metalness texture map."""
        return self._metalness_map

    @property
    def normal_map(self):
        """The normal map texture."""
        return self._normal_map

    @property
    def normal_scale(self):
        """The normal sale."""
        return self._normal_scale

    @property
    def refraction_ratio(self):
        """The refraction ratio."""
        return self._refraction_ratio

    @property
    def roughness(self):
        """The roughness intensity."""
        return self._roughness

    @property
    def roughness_map(self):
        """The roughness texture map."""
        return self._roughness_map

    @property
    def clearcoat(self):
        """Represents the intensity of the clear coat layer."""
        return self._clearcoat

    @property
    def clearcoat_map(self):
        """Clear coat texture map."""
        return self._clearcoat_map

    @property
    def clearcoat_normal_map(self):
        """Clear coat normal texture map."""
        return self._clearcoat_normal_map

    @property
    def clearcoat_normal_scale(self):
        """Clear coat normal scale."""
        return self._clearcoat_normal_scale

    @property
    def clearcoat_roughness(self):
        """The clear coat roughness."""
        return self._clearcoat_roughness

    @property
    def clearcoat_roughness_map(self):
        """The clear coat roughness texture map."""
        return self._clearcoat_roughness_map

    @property
    def ior(self):
        """Index of refraction of the material"""
        return self._ior

    @property
    def reflectivity(self):
        """The reflectivity of the material."""
        return self._reflectivity

    @property
    def sheen(self):
        """The intensity of the sheen layer."""
        return self._sheen

    @property
    def sheen_roughness(self):
        """The sheen layer roughness."""
        return self._sheen_roughness

    @property
    def sheen_roughness_map(self):
        """The sheen layer roughness texture map."""
        return self._sheen_roughness_map

    @property
    def sheen_color(self):
        """The sheen color."""
        return self._sheen_color

    @property
    def sheen_color_map(self):
        """The sheen color texture map."""
        return self._sheen_color_map

    @property
    def specular_intensity(self):
        """The specular intensity."""
        return self._specular_intensity

    @property
    def specular_intensity_map(self):
        """The specular intensity texture map."""
        return self._specular_intensity_map

    @property
    def specular_color(self):
        """The specular color."""
        return self._specular_color

    @property
    def specular_color_map(self):
        """The specular color texture map."""
        return self._specular_color_map

    @property
    def transmission(self):
        """The material transmission."""
        return self._transmission

    @property
    def transmission_map(self):
        """The material transmission texture map."""
        return self._transmission_map

    #######################################
    # Constructor
    #######################################

    def __init__(self, opacity=1, shadow_side="null", side="null", alpha_map="", ao_map="", \
                 ao_map_intensity=1, bump_map="", bump_map_scale=1, color="white", emissive="black", emissive_map="", \
                 emissive_intensity=1, env_map="", env_map_intensity=1, flat_shading=False, light_map="", \
                 light_map_itensity=1, texture_map="", metalness=0, metalness_map="", normal_map="", \
                 normal_scale=[1, 1], refraction_ratio=0.98, roughness=1, roughness_map="", clearcoat=0, \
                 clearcoat_map="", clearcoat_normal_map="", clearcoat_normal_scale=[1, 1], clearcoat_roughness=0, \
                 clearcoat_roughness_map="", ior=1.5, reflectivity=0.5, sheen=0, sheen_roughness=1,
                 sheen_roughness_map="", \
                 sheen_color="white", sheen_color_map="", specular_intensity=0, specular_intensity_map="",
                 specular_color="white", \
                 specular_color_map="", transmission=0, transmission_map=""):

        # Error handling

        if not Utils.is_a_number(opacity) or opacity < 0 or opacity > 1:
            raise Exception(
                "The parameter 'opacity' should be a float between 0 (fully transparent) and 1 (fully opaque).")

        if not (str(type(shadow_side)) == "<class 'str'>") or not ( \
                        (shadow_side == "FrontSide") or \
                        (shadow_side == "BackSide") or \
                        (shadow_side == "DoubleSide") or \
                        (shadow_side == "null")):
            raise Exception(
                "The parameter 'shadow_side' should be either \"FrontSide\", \"BackSide\", \"DoubleSide\" or \"null\".")

        if not (str(type(side)) == "<class 'str'>") or not ( \
                        (side == "FrontSide") or \
                        (side == "BackSide") or \
                        (side == "DoubleSide") or \
                        (side == "null")):
            raise Exception(
                "The parameter 'side' should be either \"FrontSide\", \"BackSide\" or \"DoubleSide\" or \"null\".")

        image_types = ["png", "bmp", "jpg", "jpeg"]

        if not (alpha_map == "" or Utils.get_uaibot_type(alpha_map) == "uaibot.Texture"):
            alpha_map = Texture(alpha_map)

        if not (ao_map == "" or Utils.get_uaibot_type(ao_map) == "uaibot.Texture"):
            ao_map = Texture(ao_map)

        if not Utils.is_a_number(ao_map_intensity) or ao_map_intensity < 0 or ao_map_intensity > 1:
            raise Exception(
                "The parameter 'ao_map_intensity' should be a float between 0 (no occlusion) and 1 (occlusion).")

        if not (bump_map == "" or Utils.get_uaibot_type(bump_map) == "uaibot.Texture"):
            bump_map = Texture(bump_map)

        if not Utils.is_a_number(bump_map_scale) or bump_map_scale < 0:
            raise Exception("The parameter 'bump_map_scale' should be a float greater or equal than 0.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a HTML-style color.")

        if not Utils.is_a_color(emissive):
            raise Exception("The parameter 'emissive' should be a HTML-style color.")

        if not (emissive_map == "" or Utils.get_uaibot_type(emissive_map) == "uaibot.Texture"):
            emissive_map = Texture(emissive_map)

        if not Utils.is_a_number(emissive_intensity) or emissive_intensity < 0:
            raise Exception("The parameter 'emissive_intensity' should be a float greater or equal than 0.")

        if not (env_map == "" or Utils.get_uaibot_type(env_map) == "uaibot.Texture"):
            env_map = Texture(env_map)

        if not Utils.is_a_number(env_map_intensity) or env_map_intensity < 0:
            raise Exception("The parameter 'env_map_intensity' should be a float greater or equal than 0.")

        if not (str(type(flat_shading)) == "<class 'bool'>"):
            raise Exception("The parameter '_flat_shading' should be a boolean.")

        if not (light_map == "" or Utils.get_uaibot_type(light_map) == "uaibot.Texture"):
            light_map = Texture(light_map)

        if not Utils.is_a_number(light_map_itensity) or light_map_itensity < 0:
            raise Exception("The parameter 'light_map_itensity' should be a float greater or equal than 0.")

        if not (texture_map == "" or Utils.get_uaibot_type(texture_map) == "uaibot.Texture"):
            texture_map = Texture(texture_map)

        if not Utils.is_a_number(metalness) or metalness < 0 or metalness > 1:
            raise Exception("The parameter 'metalness' should be a float between 0 (non-metal) and 1 (metal).")

        if not (metalness_map == "" or Utils.get_uaibot_type(metalness_map) == "uaibot.Texture"):
            metalness_map = Texture(metalness_map)

        if not (normal_map == "" or Utils.get_uaibot_type(normal_map) == "uaibot.Texture"):
            normal_map = Texture(normal_map)

        if not Utils.is_a_vector(normal_scale, 2):
            raise Exception("The parameter 'normal_scale' should be a 2D vector.")

        if not Utils.is_a_number(refraction_ratio) or refraction_ratio < 0 or refraction_ratio > 1:
            raise Exception("The parameter 'refraction_ratio' should be a float between 0 and 1.")

        if not Utils.is_a_number(roughness) or roughness < 0 or roughness > 1:
            raise Exception("The parameter 'roughness' should be a float between 0 (mirror) and 1 (fully diffuse).")

        if not (roughness_map == "" or Utils.get_uaibot_type(roughness_map) == "uaibot.Texture"):
            roughness_map = Texture(roughness_map)

        if not Utils.is_a_number(clearcoat) or clearcoat < 0 or clearcoat > 1:
            raise Exception("The parameter 'clearcoat' should be a float between 0 and 1.")

        if not (clearcoat_map == "" or Utils.get_uaibot_type(clearcoat_map) == "uaibot.Texture"):
            clearcoat_map = Texture(clearcoat_map)

        if not (clearcoat_normal_map == "" or Utils.get_uaibot_type(clearcoat_normal_map) == "uaibot.Texture"):
            clearcoat_normal_map = Texture(clearcoat_normal_map)

        if not Utils.is_a_vector(clearcoat_normal_scale, 2):
            raise Exception("The parameter 'clearcoat_normal_scale' should be a 2D vector.")

        if not Utils.is_a_number(clearcoat_roughness) or clearcoat_roughness < 0 or clearcoat_roughness > 1:
            raise Exception("The parameter 'clearcoat_roughness' should be a float between 0 and 1.")

        if not (clearcoat_roughness_map == "" or Utils.get_uaibot_type(
                clearcoat_roughness_map) == "uaibot.Texture"):
            clearcoat_roughness_map = Texture(clearcoat_roughness_map)

        if not Utils.is_a_number(ior) or ior < 1 or ior > 2.333:
            raise Exception("The parameter 'ior' should be a float between 1 and 2.333.")

        if not Utils.is_a_number(reflectivity) or reflectivity < 0 or reflectivity > 1:
            raise Exception("The parameter 'reflectivity' should be a float between 0 and 1.")

        if not Utils.is_a_number(sheen) or sheen < 0 or sheen > 1:
            raise Exception("The parameter 'sheen' should be a float between 0 and 1.")

        if not Utils.is_a_number(sheen_roughness) or sheen_roughness < 0 or sheen_roughness > 1:
            raise Exception("The parameter 'sheen_roughness' should be a float between 0 and 1.")

        if not (sheen_roughness_map == "" or Utils.get_uaibot_type(sheen_roughness_map) == "uaibot.Texture"):
            sheen_roughness_map = Texture(sheen_roughness_map)

        if not Utils.is_a_color(sheen_color):
            raise Exception("The parameter 'sheen_color' should be a HTML-style color.")

        if not (sheen_color_map == "" or Utils.get_uaibot_type(sheen_color_map) == "uaibot.Texture"):
            sheen_color_map = Texture(sheen_color_map)

        if not Utils.is_a_number(specular_intensity) or specular_intensity < 0 or specular_intensity > 1:
            raise Exception("The parameter 'specular_intensity' should be a float between 0 and 1.")

        if not (specular_intensity_map == "" or Utils.get_uaibot_type(
                specular_intensity_map) == "uaibot.Texture"):
            specular_intensity_map = Texture(specular_intensity_map)

        if not Utils.is_a_color(specular_color):
            raise Exception("The parameter 'specular_color' should be a HTML-style color.")

        if not (specular_color_map == "" or Utils.get_uaibot_type(specular_color_map) == "uaibot.Texture"):
            specular_color_map = Texture(specular_color_map)

        if not Utils.is_a_number(transmission) or transmission < 0 or transmission > 1:
            raise Exception("The parameter 'transmission' should be a float between 0 and 1.")

        if not (transmission_map == "" or Utils.get_uaibot_type(transmission_map) == "uaibot.Texture"):
            transmission_map = Texture(transmission_map)

        if transmission > 0 and opacity < 1:
            raise Exception("If 'transmission' is non-zero, 'opacity' should be set to 1.")
        # Code

        self._opacity = opacity
        self._shadow_side = shadow_side
        self._side = side

        self._transparent = self._opacity < 1
        self._alpha_map = alpha_map
        self._ao_map = ao_map
        self._ao_map_intensity = ao_map_intensity
        self._bump_map = bump_map
        self._bump_map_scale = bump_map_scale
        self._color = color
        self._emissive = emissive
        self._emissive_map = emissive_map
        self._emissive_intensity = emissive_intensity
        self._env_map = env_map
        self._env_map_intensity = env_map_intensity
        self._flat_shading = flat_shading
        self._light_map = light_map
        self._light_map_itensity = light_map_itensity
        self._map = texture_map
        self._metalness = metalness
        self._metalness_map = metalness_map
        self._normal_map = normal_map
        self._normal_scale = normal_scale
        self._refraction_ratio = refraction_ratio
        self._roughness = roughness
        self._roughness_map = roughness_map

        self._clearcoat = clearcoat
        self._clearcoat_map = clearcoat_map
        self._clearcoat_normal_map = clearcoat_normal_map
        self._clearcoat_normal_scale = clearcoat_normal_scale
        self._clearcoat_roughness = clearcoat_roughness
        self._clearcoat_roughness_map = clearcoat_roughness_map
        self._ior = ior
        self._reflectivity = reflectivity
        self._sheen = sheen
        self._sheen_roughness = sheen_roughness
        self._sheen_roughness_map = sheen_roughness_map
        self._sheen_color = sheen_color
        self._sheen_color_map = sheen_color_map
        self._specular_intensity = specular_intensity
        self._specular_intensity_map = specular_intensity_map
        self._specular_color = specular_color
        self._specular_color_map = specular_color_map
        self._transmission = transmission
        self._transmission_map = transmission_map

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Mesh Material"

        return string

    #######################################
    # Methods
    #######################################


    @staticmethod
    def create_glass(color : Optional[str] = None)-> "MeshMaterial":
        
        if color is None:
            return MeshMaterial(metalness = 0.9, roughness = 0.05, clearcoat = 1,
                            reflectivity = 0.2,  env_map_intensity=  0.9, side='DoubleSide',
                            refraction_ratio= 0.985, ior = 1, opacity=0.5)
        else:
            return MeshMaterial(metalness = 0.9, roughness = 0.05, clearcoat = 1,
                            reflectivity = 0.2,  env_map_intensity=  0.9, side='DoubleSide',
                            refraction_ratio= 0.985, ior = 1, opacity=0.5, color = color)            
 
    @staticmethod
    def create_colored_metal(color : str = 'white')-> "MeshMaterial":
   
        texture = Texture(
            url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/white.png',
            wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[4, 4])

        return MeshMaterial(metalness=0.9, clearcoat=0.5, roughness=0.0, 
                            flat_shading=True, normal_scale=[1, 1], 
                            texture_map=texture, color=color)


    @staticmethod
    def create_gold()-> "MeshMaterial":
   
        texture = Texture(
            url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/gold_metal.png',
            wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[4, 4])

        return MeshMaterial(metalness=0.9, clearcoat=0.5, roughness=0.0, 
                            flat_shading=True, normal_scale=[1, 1], 
                            texture_map=texture)

    @staticmethod
    def create_wood()-> "MeshMaterial":
   
        texture = Texture(
            url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/wood_1.jpg',
            wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[4, 4])

        return MeshMaterial(metalness=0, roughness=1.0, 
                            flat_shading=True, normal_scale=[1, 1], 
                            texture_map=texture)

    @staticmethod
    def create_rough_metal()-> "MeshMaterial":
   
        texture = Texture(
            url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/Textures/rough_metal.jpg',
            wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[4, 4])

        return MeshMaterial(metalness=0.5, roughness=0.5, 
                            flat_shading=True, normal_scale=[1, 1], 
                            texture_map=texture)
                                                 
    def gen_code(self, name):

        string = ""

        # Create textures
        if not (self.map == ""):
            string += self.map.gen_code("map_" + name)

        if not (self.alpha_map == ""):
            string += self.alpha_map.gen_code("alpha_map_" + name)

        if not (self.ao_map == ""):
            string += self.ao_map.gen_code("ao_map_" + name)

        if not (self.bump_map == ""):
            string += self.bump_map.gen_code("bump_map_" + name)

        if not (self.emissive_map == ""):
            string += self.emissive_map.gen_code("emissive_map_" + name)

        if not (self.env_map == ""):
            string += self.env_map.gen_code("env_map_" + name)

        if not (self.light_map == ""):
            string += self.light_map.gen_code("light_map_" + name)

        if not (self.metalness_map == ""):
            string += self.metalness_map.gen_code("metalness_map_" + name)

        if not (self.normal_map == ""):
            string += self.normal_map.gen_code("normal_map_" + name)

        if not (self.roughness_map == ""):
            string += self.roughness_map.gen_code("roughness_map_" + name)

        if not (self.clearcoat_map == ""):
            string += self.clearcoat_map.gen_code("clearcoat_map_" + name)

        if not (self.clearcoat_normal_map == ""):
            string += self.clearcoat_normal_map.gen_code("clearcoat_normal_map_" + name)

        if not (self.clearcoat_roughness_map == ""):
            string += self.clearcoat_roughness_map.gen_code("clearcoat_roughness_map_" + name)

        if not (self.sheen_roughness_map == ""):
            string += self.sheen_roughness_map.gen_code("sheen_roughness_map_" + name)

        if not (self.sheen_color_map == ""):
            string += self.sheen_color_map.gen_code("sheen_color_map_" + name)

        if not (self.specular_intensity_map == ""):
            string += self.specular_intensity_map.gen_code("specular_intensity_map_" + name)

        if not (self.specular_color_map == ""):
            string += self.specular_color_map.gen_code("specular_color_map_" + name)

        if not (self.transmission_map == ""):
            string += self.transmission_map.gen_code("transmission_map_" + name)

        # Create mesh
        string += "const material_" + name + " = new MeshPhysicalMaterial({\n"

        string += "opacity: " + str(self.opacity) + ",\n"

        if not (self.shadow_side == "null"):
            string += "shadowSide: " + self.shadow_side + ",\n"

        if not (self.side == "null"):
            string += "side: " + self.side + ",\n"

        string += "transparent: " + str(self.transparent).lower() + ",\n"

        if not (self.alpha_map == ""):
            string += "alphaMap: texture_alpha_map_" + name + ",\n"

        if not (self.ao_map == ""):
            string += "aoMap: texture_alpha_map_" + name + ",\n"

        string += "aoMapIntensity: " + str(self.ao_map_intensity) + ",\n"

        if not (self.bump_map == ""):
            string += "bumpMap: texture_bump_map_" + name + ",\n"

        string += "bumpScale: " + str(self.bump_map_scale) + ",\n"
        string += "color: \"" + self.color + "\",\n"
        string += "emissive: \"" + self.emissive + "\",\n"

        if not (self.emissive_map == ""):
            string += "emissiveMap: texture_emissive_map_" + name + ",\n"

        string += "emissiveIntensity: " + str(self.emissive_intensity) + ",\n"

        if not (self.env_map == ""):
            string += "envMap: texture_env_map_" + name + ",\n"

        string += "envMapIntensity: " + str(self.env_map_intensity) + ",\n"
        string += "flatShading: " + str(self.flat_shading).lower() + ",\n"

        if not (self.light_map == ""):
            string += "lightMap: texture_light_map_" + name + ",\n"

        string += "lightMapIntensity: " + str(self.light_map_itensity) + ",\n"

        if not (self.map == ""):
            string += "map: texture_map_" + name + ",\n"

        string += "metalness: " + str(self.metalness) + ",\n"

        if not (self.metalness_map == ""):
            string += "metalnessMap: texture_metalness_map_" + name + ",\n"

        if not (self.normal_map == ""):
            string += "normalMap: texture_normal_map_" + name + ",\n"

        string += "normalScale: new Vector2(" + str(self.normal_scale[0]) + "," + str(self.normal_scale[1]) + "),\n"
        string += "refractionRatio: " + str(self.refraction_ratio) + ",\n"
        string += "roughness: " + str(self.roughness) + ",\n"

        if not (self.roughness_map == ""):
            string += "roughnessMap: texture_roughness_map_" + name + ",\n"

        string += "clearcoat: " + str(self.clearcoat) + ",\n"

        if not (self.clearcoat_map == ""):
            string += "clearcoatMap: texture_clearcoat_map_" + name + ",\n"

        if not (self.clearcoat_normal_map == ""):
            string += "clearcoatNormalMap: texture_clearcoat_normal_map_" + name + ",\n"

        string += "clearcoatNormalScale: new Vector2(" + str(self.clearcoat_normal_scale[0]) + "," + str(
            self.clearcoat_normal_scale[1]) + "),\n"
        string += "clearcoatRoughness: " + str(self.clearcoat_roughness) + ",\n"

        if not (self.clearcoat_roughness_map == ""):
            string += "clearcoatRoughnessMap: texture_clearcoat_roughness_map_" + name + ",\n"

        string += "ior: " + str(self.ior) + ",\n"
        string += "reflectivity: " + str(self.reflectivity) + ",\n"
        string += "sheen: " + str(self.sheen) + ",\n"
        string += "sheenRoughness: " + str(self.sheen_roughness) + ",\n"

        if not (self.sheen_roughness_map == ""):
            string += "sheenRoughnessMap: texture_sheen_roughness_map_" + name + ",\n"

        string += "sheenColor: \"" + self.sheen_color + "\",\n"

        if not (self.sheen_color_map == ""):
            string += "sheenColorMap: texture_sheen_color_map" + name + ",\n"

        string += "specularIntensity: " + str(self.specular_intensity) + ",\n"

        if not (self.specular_intensity_map == ""):
            string += "specularIntensityMap: texture_specular_intensity_map_" + name + ",\n"

        string += "specularColor: \"" + self._specular_color + "\",\n"

        if not (self.specular_color_map == ""):
            string += "specularColorMap: texture_specular_color_map_" + name + ",\n"

        string += "transmission: " + str(self._transmission) + ",\n"

        if not (self.transmission_map == ""):
            string += "transmissionMap: texture_transmission_map_" + name + ",\n"

        string += "});\n\n"

        return string



from utils import *
import numpy as np
import os

# Geometric Jacobian
def _jac_geo(self, q, axis, htm, mode):
    if q is None:
        q = self.q
    if htm is None:
        htm = np.matrix(self.htm)

    n = len(self.links)

    # Error handling
    if mode not in ['python','c++','auto']:
        raise Exception("The parameter 'mode' should be 'python,'c++', or 'auto'.")
        
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not (axis == "eef" or axis == "dh" or axis == "com"):
        raise Exception("The parameter 'axis' should be one of the following strings:\n" \
                        "'eef': End-effector \n" \
                        "'dh': All " + str(n) + " axis of Denavit-Hartenberg\n" \
                                                "'com': All " + str(
            n) + " axis centered at the center of mass of the objects.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")
    if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
        raise Exception("c++ mode is set, but .so file was not loaded!")
    # end error handling

    if mode == 'python' or axis == 'com' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
        return _jac_geo_python(self, Utils.cvt(q), axis, htm)
    else:
        fk_res = self.cpp_robot.fk(Utils.cvt(q), htm, True)

        if axis=='eef':
            return Utils.cvt(np.vstack((fk_res.jac_v_ee, fk_res.jac_w_ee))) , Utils.cvt(fk_res.htm_ee)
        else:
            htm_dh = []
            jac_dh = []
            for i in range(n):
                htm_dh.append(Utils.cvt(fk_res.htm_dh[i]))
                jac_dh.append(Utils.cvt(np.vstack((fk_res.jac_v_dh[i], fk_res.jac_w_dh[i]))))

            return jac_dh, htm_dh

def _jac_geo_python(self, q=None, axis='eef', htm=None):

    n = len(self.links)


    if axis == 'dh' or axis == 'eef':
        htm_for_jac = self.fkm(q, 'dh', htm, mode='python')
    if axis == 'com':
        htm_for_jac = self.fkm(q, 'com', htm, mode='python')

    jac = [np.matrix(np.zeros((6,n))) for i in range(n)]

    htm_world_0 = htm * self.htm_base_0

    for i in range(n):
        p_i = htm_for_jac[i][0:3, 3]
        for j in range(i + 1):

            if j == 0:
                p_j_ant = htm_world_0[0:3, 3]
                z_j_ant = htm_world_0[0:3, 2]
            else:
                p_j_ant = htm_for_jac[j - 1][0:3, 3]
                z_j_ant = htm_for_jac[j - 1][0:3, 2]

            if self.links[j].joint_type == 0:
                jac[i][0:3, j] = Utils.S(z_j_ant) * (p_i - p_j_ant)
                jac[i][3:6, j] = z_j_ant

            if self.links[j].joint_type == 1:
                jac[i][0:3, j] = z_j_ant
                jac[i][3:6, j] = np.matrix(np.zeros((3,1)))

    if axis == 'dh' or axis == 'com':
        return jac, htm_for_jac

    if axis == 'eef':
        htm_0_eef =  htm_for_jac[-1][:, :] * self.htm_n_eef
        jg = jac[-1][:, :]
        jv = jg[0:3,:]
        jw = jg[3:6,:]
        jv = jv + Utils.S(htm_for_jac[-1][0:3,-1] - htm_0_eef[0:3,-1]) * jw
        jg = np.block([[jv],[jw]])

        return jg , htm_0_eef



from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_kuka_kr5(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    link_info = [[0, 0, 0, 0, 0, 0], #"theta" rotation in z
                 [0.335, 0.000, 0.000, -0.405, 0.000, 0.080],  # "d" translation in z
                 [-1.570, 0.000, 1.570, -1.570, -1.570, 0],  # "alfa" rotation in x
                 [0.075, 0.365, 0.090, 0.000, 0.000, 0],  # "a" translation in x
                 [0, 0, 0, 0, 0, 0]] #joint type

    n = 6

    # Collision model
    col_model = [[], [], [], [], [], []]

    col_model[0].append(Cylinder(htm=Utils.trn([-0.075, 0.170, 0]) * Utils.roty(3.14 / 2) * Utils.rotx(3.14 / 2),
                                 name=name + "_C0_0", radius=0.12, height=0.33, color="red", opacity=0.3))

    col_model[0].append(Cylinder(htm=Utils.trn([0, 0, 0.030]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14),
                                 name=name + "_C0_1", radius=0.095, height=0.30, color="red", opacity=0.3))

    col_model[1].append(Box(htm=Utils.trn([-0.200, 0.020, 0.12]) * Utils.roty(3.14 / 2),
                            name=name + "_C1_0", width=0.1, height=0.5, depth=0.16, color="green", opacity=0.3))

    col_model[1].append(Cylinder(htm=Utils.trn([0, 0, 0.040]) * Utils.rotz(3.14) * Utils.rotx(3.14),
                                 name=name + "_C1_1", radius=0.095, height=0.28, color="green", opacity=0.3))

    col_model[2].append(Box(htm=Utils.trn([0, 0, -0.224]) * Utils.rotz(-3.14 / 2) * Utils.rotx(-3.14 / 2),
                            name=name + "_C2_0", width=0.143, height=0.12, depth=0.45, color="blue", opacity=0.3))

    # Create 3d objects
    htm1 = np.matrix([[-0.0024, 0.0008, -1., 0.], [-1., -0., 0.0024, 0.], [0., 1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm2 = np.matrix([[1., 0., 0., -0.08], [0., -1., -0.0016, 0.135], [0., 0.0016, -1., 0.], [0., 0., 0., 1.]])
    htm3 = np.matrix([[0.24, 0.0008, -0.9708, -0.37], [-0.9708, 0.0002, -0.24, 0.], [0., 1., 0.0008, 0.11], [0., 0., 0., 1.]])
    htm4 = np.matrix([[0.0008, 1., 0., -0.09], [-1., 0.0008, 0., 0.], [0., 0., 1., 0.], [0., 0., 0., 1.]])
    htm5 = np.matrix([[0.0008, -1., 0., 0.], [1., 0.0008, 0., -0.185], [0., 0., 1., 0.], [0., 0., 0., 1.]])
    htm6 = np.matrix([[-1., 0., 0.0016, 0.], [0., 1., 0., 0.], [-0.0016, 0., -1., 0.], [0., 0., 0., 1.]])
    htm7 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, -1., 0.], [0., 1., 0.0008, -0.012], [0., 0., 0., 1.]])
    
    base_3d_obj = [
        Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaKR5/Base.obj',
                0.001,
                htm1,
                MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#242526",
                             opacity=opacity))]

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaKR5/Axis1.obj',
                 0.001,
                 htm2,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaKR5/Axis2.obj',
                 0.001,
                 htm3,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaKR5/Axis3.obj',
                 0.001,
                 htm4,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaKR5/Axis4.obj',
                 0.001,
                 htm5,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaKR5/Axis5.obj',
                 0.001,
                 htm6,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaKR5/Axis6.obj',
                 0.001,
                 htm7,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#242526",
                              opacity=opacity))]
    )

    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    q0 = [1.570, -1.570, 0.000, 0.000, 0, 0.000]

    #Create joint limits
    joint_limits = (np.pi/180)*np.matrix([[-170,170],[-190, 45],[-119-90, 165-90],[-190,190],[-120,120],[-358,358]])

    return base_3d_obj, links, np.identity(4), np.identity(4), q0, joint_limits



import numpy as np
import os
from utils import Utils

if os.environ['CPP_SO_FOUND'] == "1":
    import uaibot_cpp_bind as ub_cpp

_DELTA = 1e-3
_DS = 1e-3

def _vector_field_SE3(state, curve, kt1=1.0, kt2=1.0, kt3=1.0, kn1=1.0, kn2=1.0, curve_derivative=[], delta=_DELTA, ds=_DS, mode='auto'):
    n = 4
    # Error handling
    if mode not in ['python', 'c++', 'auto']:
        raise Exception("The parameter 'mode' should be 'python,'c++', or 'auto'.")

    if not Utils.is_a_matrix(state, n, n):
        raise Exception(f"The parameter 'state' should be a {n}x{n} matrix.")
    if not Utils.is_a_number(kt1):
        raise Exception("The parameter 'kt1' should be a float.")
    if not Utils.is_a_number(kt2):
        raise Exception("The parameter 'kt2' should be a float.")
    if not Utils.is_a_number(kt3):
        raise Exception("The parameter 'kt3' should be a float.")
    if not Utils.is_a_number(kn1):
        raise Exception("The parameter 'kn1' should be a float.")
    if not Utils.is_a_number(kn2):
        raise Exception("The parameter 'kn2' should be a float.")
    if not Utils.is_a_number(delta) or delta <= 0:
        raise Exception("The parameter 'delta' should be a positive float.")
    if not Utils.is_a_number(ds) or ds <= 0:
        raise Exception("The parameter 'delta' should be a positive float.")

    if mode == 'c++' and os.environ['CPP_SO_FOUND'] == '0':
        raise ImportError("c++ mode is set, but .so file was not loaded!")
    if mode == 'python' or (mode == 'auto' and os.environ['CPP_SO_FOUND'] == '0'):
        vf_res = _vector_field_SE3_python(state, curve, kt1, kt2, kt3, kn1, kn2, curve_derivative, delta, ds)
        return None, None, None
    else:
        vf_res = ub_cpp.vectorfield_SE3(state, curve, kt1, kt2, kt3, kn1, kn2, curve_derivative, delta, ds)
        return vf_res.twist, vf_res.dist, vf_res.index

def _vector_field_SE3_python(pstate, curve, kt1=1.0, kt2=1.0, kt3=1.0, kn1=1.0, kn2=1.0, curve_derivative=[], delta=_DELTA, ds=_DS):
    raise NotImplementedError("The python-only version has not been implemented yet.")
    return None




from utils import *


class HTMLDiv:
    """
  A HTML div that is used to display a text into the scenario canvas.

  Parameters
  ----------
  name : string
      The object's name.
      (default: '' (automatic)).

  html_text : string
      The html string.
      (default: 'Text').

  style : string
      CSS style string  (example: 'position:absolute;top:200px,right:100px').
      (default: 'Text').
  """

    #######################################
    # Attributes
    #######################################

    @property
    def name(self) -> str:
        """The name of the object."""
        return self._name

    @property
    def html_text(self) -> str:
        """The HTML string representing the text. HTML markup is allowed here."""
        return self._html_text

    @property
    def style(self) -> str:
        """The CSS style string."""
        return self._style

    #######################################
    # Constructor
    #######################################

    def __init__(self, name: str ="", html_text: str ="Text", style: str ="")-> "HTMLDiv":

        # Error handling

        if name=="":
            name="var_htmldiv_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. "
                "It should not begin with a number.")

        if not str(type(html_text)) == "<class 'str'>":
            raise Exception("The parameter 'html_text' should be a string.")

        if not str(type(style)) == "<class 'str'>":
            raise Exception("The parameter 'style' should be a string.")

        # end error handling

        self._name = name
        self._html_text = html_text
        self._style = style
        self._frames = []
        self._max_time = 0

        self.set_ani_frame(self.html_text)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Text with name '" + self.name + "': \n\n"
        string += " HTML string: " + str(self.html_text) + "\n"
        string += " Style: " + str(self.style) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time: float, html_text: Optional[str] =None, style: Optional[str] = None) -> None:
        """

    """
        if html_text is None:
            html_text = self.html_text
        if style is None:
            style = self.style

        # Error handling
        if not str(type(html_text)) == "<class 'str'>":
            raise Exception("The parameter 'html_text' should be either a string or 'None'.")

        if not str(type(style)) == "<class 'str'>":
            raise Exception("The parameter 'style' should be a string.")
        # end error handling

        self._html_text = html_text
        self._style = style
        self._max_time = max(self._max_time, time)

        self._frames.append([time, html_text, style])

    # Set config. Restart animation queue
    def set_ani_frame(self, html_text: Optional[str] =None, style: Optional[str] = None) -> None:
        """

    """
        if html_text is None:
            html_text = self.html_text
        if style is None:
            style = self.style

        # Error handling
        if not str(type(html_text)) == "<class 'str'>":
            raise Exception("The parameter 'html_text' should be either a string or 'None'.")

        if not str(type(style)) == "<class 'str'>":
            raise Exception("The parameter 'style' should be a string.")

        # end error handling

        self._frames = []
        self.add_ani_frame(0, html_text, style)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        string = "\n"
        string += "//BEGIN DECLARATION OF THE HTML DIV '" + self.name + "'\n\n"
        string += "const var_" + self.name + " = new HTMLDiv('" + self.name + "', " + str(self._frames) + ");\n"
        string += "sceneElements.push(var_" + self.name + ");\n"
        string += "//USER INPUT GOES HERE"

        return string



# Update collision object pose (for drawing purpose only)
def _update_col_object(self, time, mode):
    htm = self.fkm(axis="dh", mode=mode)

    for i in range(len(self.links)):
        for j in range(len(self.links[i].col_objects)):
            self.links[i].col_objects[j][0].add_ani_frame(time, htm[i] * self.links[i].col_objects[j][1])



import numpy as np
from utils import *
from graphics.texture import *


class MTLMeshMaterial:
    """
    Mesh Material loaded from a MTL file.

    Parameters
    ----------
    opacity : float between 0 and 1
        Mirrors 'Material.opacity' in threejs.
        Float in the range of 0.0 - 1.0 indicating how transparent the material is.
        A value of 0.0 indicates fully transparent, 1.0 is fully opaque.
        The attribute 'transparent' is set to true automatically if opacity<1.
        (default: 1).

    url : string
        The url that contains the mtl files.
        It must have one of the following formats: 'url'.


    """

    #######################################
    # Attributes
    #######################################



    @property
    def url(self) -> str:
        """The 3d model url."""
        return self._url

    @property
    def opacity(self) -> float:
        """The object opacity."""
        return self._opacity

    #######################################
    # Constructor
    #######################################

    def __init__(self, url: str ="", opacity: float =1) -> "MTLMeshMaterial":

        # Error handling

        if not Utils.is_a_number(opacity) or opacity < 0 or opacity > 1:
            raise Exception(
                "The parameter 'opacity' should be a float between 0 (fully transparent) and 1 (fully opaque).")

        error = Utils.is_url_available(url, ['mtl'])
        if not (error == "ok!"):
            raise Exception("The parameter 'url' " + error)
        
        self._url = url
        self._opacity = opacity


    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "MTL Mesh Material"

        return string

    #######################################
    # Methods
    #######################################

    def gen_code(self, name):

        string = "const material_"+ name + " = {"
        
        string+= "type: 'mtl',"
        string+= "opacity: "+str(self.opacity)+","
        string+= "url: '"+self.url+"'};\n\n"
        



        return string



from utils_test import *
import uaibot as ub
import time
import sys
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Qt5Agg')  
import os


print("Running Object to Robot smooth distance test")


projball1 = ub.Ball(radius=0.02,color='blue')
projball2 = ub.Ball(radius=0.02,color='red')

robot = ub.Robot.create_kuka_lbr_iiwa()

sim = ub.Simulation.create_sim_grid([projball1, projball2, robot])



sum_time = 0
max_time = 0
cross_error_max = 0
N=100

eps=0.08
h=0.1



dmin0=0.2
dmin=0.05
eta=0.5
dt=0.01
Tmax=20
Nt = round(Tmax/dt)
No = 10
counter=0


der_ana_hist_all = []
der_num_hist_all = []
name_hist = []
t = 0

error_max = 0

for k in range(No):
    
    cont=True
    while cont:
        obj = generate_rand_object(color='red')
        cont = robot.compute_dist(q = robot.q, obj = obj).get_closest_item().distance<dmin0
        type = ub.Utils.get_uaibot_type(obj)    

    sim.add(obj)

    
    if type != 'uaibot.PointCloud':
        htm = np.matrix(obj.htm)
        obj.add_ani_frame(0,htm=ub.Utils.trn([1000,1000,1000]))
        obj.add_ani_frame(t,htm=htm)


    fun = lambda _q : robot.compute_dist(q = _q, obj = obj, eps=eps, h=h).get_item(5,0).distance()
    
    der_ana_hist = []
    der_num_hist = []
    dist_hist = []

    old_dr = None
    for i in range(Nt):
        t += dt
        
        qdot = np.matrix([1.0,-1.0,1.0,-1.0,1.0,-1.0,1.0]).transpose()
        robot.add_ani_frame(t, robot.q+qdot*dt)
        
        dr = robot.compute_dist(q = robot.q, obj = obj, eps=eps, h=h, no_iter_max=10, old_dist_struct = old_dr)
        old_dr = dr
        
        der_ana_hist.append(dr.get_item(5,0).distance * (dr.get_item(5,0).jac_distance*qdot)[0,0])
        dist_hist.append(0.5*(dr.get_item(5,0).distance ** 2))
        
        sys.stdout.write("\rCompleted: "+str(round(100*counter/(Nt*No)))+"%")
        sys.stdout.flush()      
        
        counter+=1

    #Analyse
    for i in range(Nt-2):
        der_num = (dist_hist[i+2]-dist_hist[i])/(2*dt)
        der_num_hist.append(der_num)
        
        error = abs(der_num - der_ana_hist[i])
        error_max = max(error, error_max)
        
    der_ana_hist_all.append(der_ana_hist[1:-1])
    der_num_hist_all.append(der_num_hist)  
    name_hist.append("Collision check with "+type) 
    
    # rr=[]
    # for i in range(len(der_num_hist)):
    #     rr.append(der_num_hist[i]/der_ana_hist[i+1])
      
    # plt.plot(rr)    
    # plt.plot(der_num_hist)
    # plt.plot(der_ana_hist[1:-1])
    # plt.show()
    
    if type != 'uaibot.PointCloud':
        obj.add_ani_frame(t,htm=ub.Utils.trn([1000,1000,1000]))
    else:
        obj.add_ani_frame(t,0,0)
            

#Plot everything

n = len(name_hist)

cols = int(np.ceil(np.sqrt(n)))
rows = int(np.ceil(n / cols))


fig, axes = plt.subplots(rows, cols, figsize=(cols * 5, rows * 4))
axes = np.array(axes).flatten()  


for i in range(n):
    ax = axes[i]
    ax.plot(der_ana_hist_all[i], label='Analytical')
    ax.plot(der_num_hist_all[i], label='Numerical', linestyle='dashed')
    ax.set_title(name_hist[i])
    ax.legend()


for j in range(n, len(axes)):
    fig.delaxes(axes[j])

plt.tight_layout()
plt.show()


print("\nMaximum error was "+str(round(error_max,3)))
print("Test was a success!")   
    
    

current_folder = os.path.dirname(os.path.abspath(__file__))
sim.save(current_folder,"test_o2o_smooth")



import numpy as np
from utils import *


class Vector:
    """
   A vector object (arrow).

   Parameters
   ----------

   name : string
       The object's name.
       (default: '' (automatic)).

   color : string
       The object's color, a HTML - compatible string.
       (default: "red").

   """

    #######################################
    # Attributes
    #######################################

    @property
    def name(self):
        """Name of the object."""
        return self._name

    @property
    def color(self):
        """Color of the object"""
        return self._color

    @property
    def thickness(self):
        """The thickness of the vetor."""
        return self._thickness

    @property
    def origin(self):
        """The origin point of the vector, in scenario coordinates and in meters."""
        return self._origin

    @property
    def vector(self):
        """The vector, in scenario coordinates and in meters."""
        return self._vector

    #######################################
    # Constructor
    #######################################

    def __init__(self, name="", color="black", thickness=1, origin=[0, 0, 0], vector=[0.5, 0.5, 0.5]):

        # Error handling

        if name=="":
            name="var_vector_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. "
                "It should not begin with a number.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a HTML-compatible color.")

        if not Utils.is_a_number(thickness) or thickness <= 0:
            raise Exception("The parameter 'thickness' should be positive float.")

        if not Utils.is_a_vector(origin, 3):
            raise Exception("The parameter 'origin' should be a 3d list or numpy array.")

        if not Utils.is_a_vector(vector, 3):
            raise Exception("The parameter 'vector' should be a 3d list or numpy array.")

        # end error handling

        self._name = name
        self._color = color
        self._thickness = thickness
        self._origin = np.matrix(origin).reshape((3,1))
        self._vector = np.matrix(vector).reshape((3,1))
        self._frames = []
        self._max_time = 0

        # Set initial total configuration
        self.set_ani_frame(self.origin, self.vector)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Vector '" + self.name + "': \n\n"
        string += "Origin: "+str(self.origin)+" \n"
        string += "Vector: " + str(self.vector) + " \n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time, origin=None, vector=None):

        if origin is None:
            origin = self.origin

        if vector is None:
            vector = self.vector

        # Error handling
        if not Utils.is_a_vector(origin, 3):
            raise Exception("The parameter 'origin' should be a 3d list or numpy array.")

        if not Utils.is_a_vector(vector, 3):
            raise Exception("The parameter 'vector' should be a 3d list or numpy array.")

        if not Utils.is_a_number(time) or time < 0:
            raise Exception("The parameter 'time' should be a positive float.")
        # end error handling

        length_vector = np.linalg.norm(np.matrix(vector))

        #if length_vector < 0.0001:
        #    raise Exception("'origin' and 'vector' are too close to each other.")

        self._origin = np.around(np.matrix(origin).reshape((3,1)),4).tolist()
        self._vector = np.matrix(vector).reshape((3,1)).tolist()
        dir = np.around((np.matrix(vector).reshape((3,1)) / (0.0001+length_vector)),4).tolist()

        self._frames.append([time, self.origin, dir, length_vector])
        self._max_time = max(self._max_time, time)

    # Set config. Restart animation queue
    def set_ani_frame(self, start_point=None, end_point=None):

        self._frames = []
        self.add_ani_frame(0, start_point, end_point)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        length_vector = np.linalg.norm(np.matrix(self.vector))
        dir = (np.matrix(self.vector) / (length_vector+0.00001)).tolist()

        string = "\n"
        string += "//BEGIN DECLARATION OF THE VECTOR '" + self.name + "'\n\n"
        string += "const var_" + self.name + " = new Vector(" + str(self._frames) + ", '" + str(
            self.color) + "'," + str(self.thickness) + ", " + str(self.origin) + ", " + str(dir) + ", " + str(
            length_vector) + ");\n"
        string += "sceneElements.push(var_" + self.name + ");\n"
        string += "//USER INPUT GOES HERE"

        return string



import robot as rb
from utils import Utils

from simobjects.group import Group
from robot._create_davinci_arms import _create_davinci_arm1, _create_davinci_arm2, _create_davinci_arm3, _create_davinci_arm4
from robot._create_davinci_chest import _create_davinci_chest
import numpy as np


def _create_davinci(htm, name, color='#3e3f42', opacity=1, eef_frame_visible=True, joint_limits=None):

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception(
            "The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception(
            "The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception(
            "The parameter 'opacity' should be a float between 0 and 1.")

    arm1_links, arm1_base_3d_obj, arm1_htmbase, arm1_htmeef, arm1_q0, arm1_limits = _create_davinci_arm1(
        color=color, opacity=opacity, name=name + '_arm_1')
    arm2_links, arm2_base_3d_obj, arm2_htmbase, arm2_htmeef, arm2_q0, arm2_limits = _create_davinci_arm2(
        color=color, opacity=opacity, name=name + '_arm_2')
    arm3_links, arm3_base_3d_obj, arm3_htmbase, arm3_htmeef, arm3_q0, arm3_limits = _create_davinci_arm3(
        color=color, opacity=opacity, name=name + '_arm_3')
    arm4_links, arm4_base_3d_obj, arm4_htmbase, arm4_htmeef, arm4_q0, arm4_limits = _create_davinci_arm4(
        color=color, opacity=opacity, name=name + '_arm_4')
    chest = _create_davinci_chest(name=name, color=color, opacity=opacity)

    robot_arm1 = rb.Robot(name + "__arm1", links=arm1_links, list_base_3d_obj=arm1_base_3d_obj, htm=np.identity(4),
                          htm_base_0=arm1_htmbase, htm_n_eef=arm1_htmeef, q0=arm1_q0, eef_frame_visible=eef_frame_visible, joint_limits=arm1_limits)
    robot_arm2 = rb.Robot(name + "__arm2", links=arm2_links, list_base_3d_obj=arm2_base_3d_obj, htm=np.identity(4),
                          htm_base_0=arm2_htmbase, htm_n_eef=arm2_htmeef, q0=arm2_q0, eef_frame_visible=eef_frame_visible, joint_limits=arm2_limits)
    robot_arm3 = rb.Robot(name + "__arm3", links=arm3_links, list_base_3d_obj=arm3_base_3d_obj, htm=np.identity(4),
                          htm_base_0=arm3_htmbase, htm_n_eef=arm3_htmeef, q0=arm3_q0, eef_frame_visible=eef_frame_visible, joint_limits=arm3_limits)
    robot_arm4 = rb.Robot(name + "__arm4", links=arm4_links, list_base_3d_obj=arm4_base_3d_obj, htm=np.identity(4),
                          htm_base_0=arm4_htmbase, htm_n_eef=arm4_htmeef, q0=arm4_q0, eef_frame_visible=eef_frame_visible, joint_limits=arm4_limits)
    return Group([robot_arm1, robot_arm2, robot_arm3, robot_arm4, chest])



import uaibot as ub
import numpy as np
from utils_test import *
# from uaibot.utils import Utils



input1 = [1,2,3]
input2 = [[1],[2],[3]]
input3 = np.matrix([1,2,3])
input4 = np.array([1,2,3])
input5 = np.matrix([1,2,3]).T


print(ub.Utils.S(input1))
print(ub.Utils.S(input2))
print(ub.Utils.S(input3))
print(ub.Utils.S(input4))
print(ub.Utils.S(input5))

print(ub.Utils.rot(input1,3))
print(ub.Utils.rot(input2,3))
print(ub.Utils.rot(input3,3))
print(ub.Utils.rot(input4,3))
print(ub.Utils.rot(input5,3))


print(ub.Utils.trn(input2))
print(ub.Utils.trn(input3))
print(ub.Utils.trn(input4))
print(ub.Utils.trn(input5))

print(ub.Utils.inv_htm(ub.Utils.rot(input1,3)*ub.Utils.trn([0,1,2])))
print(ub.Utils.inv_htm(ub.Utils.rot(input2,3)*ub.Utils.trn([0,1,2])))
print(ub.Utils.inv_htm(ub.Utils.rot(input3,3)*ub.Utils.trn([0,1,2])))
print(ub.Utils.inv_htm(ub.Utils.rot(input4,3)*ub.Utils.trn([0,1,2])))
print(ub.Utils.inv_htm(ub.Utils.rot(input5,3)*ub.Utils.trn([0,1,2])))

a, b = ub.Utils.axis_angle(ub.Utils.inv_htm(ub.Utils.rot(input1,3)*ub.Utils.trn([0,1,2])))

print(a)
print(b)

a, b, c = ub.Utils.euler_angles(ub.Utils.inv_htm(ub.Utils.rot(input1,3)*ub.Utils.trn([0,1,2])))

print([a,b,c])

A1 = [[-1,-2,3],[4,5,6]]
A2 = np.matrix([[-1,-2,3],[4,5,6]])
A3 = np.array([[-1,-2,3],[4,5,6]])

b1 = [1,0]
b2 = np.matrix([1,0])
b3 = np.matrix([1,0]).T

print(ub.Utils.dp_inv(A1))
print(ub.Utils.dp_inv(A2))
print(ub.Utils.dp_inv(A3))

print(ub.Utils.dp_inv_solve(A1, b1))
print(ub.Utils.dp_inv_solve(A2, b2))
print(ub.Utils.dp_inv_solve(A3, b3))
print(ub.Utils.dp_inv_solve(A1, b1, mode='python'))
print(ub.Utils.dp_inv_solve(A2, b2, mode='python'))
print(ub.Utils.dp_inv_solve(A3, b3, mode='python'))

H1 = [[2,1],[1,3]]
H2 = np.array(H1)
H3 = np.matrix(H1)
H4 = np.matrix(H1)

f1 = [1,2]
f2 = np.matrix([1,2])
f3 = np.matrix([1,2]).T
f4 = np.array([1,2])

A1 = [[1,2],[3,4],[4,5],[5,6]]
A2 = np.matrix([[1,2],[3,4],[4,5],[5,6]])
A3 = np.array([[1,2],[3,4],[4,5],[5,6]])
A4 = np.array([[1,2],[3,4],[4,5],[5,6]])

b1 = [10,20,30,40]
b2 = np.matrix([10,20,30,40])
b3 = np.matrix([10,20,30,40]).T
b4 = np.array([10,20,30,40])

print("QP Solver")

print(ub.Utils.solve_qp(H1, f1, A1, b1))
print(ub.Utils.solve_qp(H2, f2, A2, b2))
print(ub.Utils.solve_qp(H3, f3, A3, b3))
print(ub.Utils.solve_qp(H4, f4, A4, b4))

robot = ub.Robot.create_franka_emika_3()

NO_TRY = 20

print("-------------------------------------------------------------")
print("Testing FK: ")

max_error = 0
for i in range(NO_TRY):
    q_input = [np.random.randn() for j in range(7)]
    
    fk1 = robot.fkm(q=q_input,mode='c++')
    fk2 = robot.fkm(q=np.matrix(q_input),mode='c++')
    fk3 = robot.fkm(q=q_input,mode='python')
    fk4 = robot.fkm(q=np.matrix(q_input),mode='python')  
    
    max_error = max(max_error, np.linalg.norm(fk1-fk2))
    max_error = max(max_error, np.linalg.norm(fk3-fk4))  
    max_error = max(max_error, np.linalg.norm(fk1-fk3))
    max_error = max(max_error, np.linalg.norm(fk2-fk4))      

print("Max error FK  = "+str(max_error))

print("-------------------------------------------------------------")
print("Testing Jacobian: ")

max_error = 0
for i in range(NO_TRY):
    q_input = [np.random.randn() for j in range(7)]
    
    jg1, fk1 = robot.jac_geo(q=q_input,mode='c++')
    jg2, fk2 = robot.jac_geo(q=np.matrix(q_input),mode='c++')
    jg3, fk3 = robot.jac_geo(q=q_input,mode='python')
    jg4, fk4 = robot.jac_geo(q=np.matrix(q_input),mode='python')  
    
    max_error = max(max_error, np.linalg.norm(fk1-fk2))
    max_error = max(max_error, np.linalg.norm(fk3-fk4))  
    max_error = max(max_error, np.linalg.norm(fk1-fk3))
    max_error = max(max_error, np.linalg.norm(fk2-fk4))      
    max_error = max(max_error, np.linalg.norm(jg1-jg2))
    max_error = max(max_error, np.linalg.norm(jg3-jg4))  
    max_error = max(max_error, np.linalg.norm(jg1-jg3))
    max_error = max(max_error, np.linalg.norm(jg2-jg4))  
    
print("Max error Jacobian  = "+str(max_error))

print("-------------------------------------------------------------")
print("Testing IK  ")

max_error = 0
no_fail = 0

obs = ub.Ball(htm = ub.Utils.trn([0.5,0,0.6]),radius=0.25)
for i in range(NO_TRY):
    q_input = np.matrix(robot.q)
    
    cont = True
    
    while cont:
        for i in range(7):
            q_input[i,0] = robot.joint_limit[i,0]+np.random.rand()*(robot.joint_limit[i,1]-robot.joint_limit[i,0])
            
        ok, _, _ = robot.check_free_config(q=q_input, obstacles=[obs])
        cont = not ok

    
    htm_rand = robot.fkm(q_input)
    
    try:
        q_inv_cpp = robot.ikm(htm_tg = htm_rand, mode='c++', obstacles=[obs])
        q_inv_py = robot.ikm(htm_tg = htm_rand, mode='python', obstacles=[obs])
        
        error_cpp = np.linalg.norm(htm_rand-robot.fkm(q_inv_cpp))
        error_py = np.linalg.norm(htm_rand-robot.fkm(q_inv_py))
        
        max_error = max(max_error, error_cpp)
        max_error = max(max_error, error_py)
    except:
        no_fail = no_fail+1
        
print("No IK fail: "+str(no_fail))
print("Max error IK = "+str(max_error))

print("-------------------------------------------------------------")
print("Testing Task Function: ")

max_error = 0
for i in range(NO_TRY):
    q_input = [np.random.randn() for j in range(7)]
    htm_rand = robot.fkm(q_input)
    q_input = [np.random.randn() for j in range(7)]
    
    r1, jac_r1 = robot.task_function(q=q_input,mode='c++', htm_tg = htm_rand)
    r2, jac_r2 = robot.task_function(q=np.matrix(q_input),mode='c++', htm_tg = htm_rand)
    r3, jac_r3 = robot.task_function(q=q_input,mode='python', htm_tg = htm_rand)
    r4, jac_r4 = robot.task_function(q=np.matrix(q_input),mode='python', htm_tg = htm_rand)  
    
    max_error = max(max_error, np.linalg.norm(r1-r2))
    max_error = max(max_error, np.linalg.norm(r3-r4))  
    max_error = max(max_error, np.linalg.norm(r1-r3))
    max_error = max(max_error, np.linalg.norm(r2-r4))      
    max_error = max(max_error, np.linalg.norm(jac_r1-jac_r2))
    max_error = max(max_error, np.linalg.norm(jac_r3-jac_r4))  
    max_error = max(max_error, np.linalg.norm(jac_r1-jac_r3))
    max_error = max(max_error, np.linalg.norm(jac_r2-jac_r4))
    
print("Max error Task Function = "+str(max_error))

print("-------------------------------------------------------------")

print("Testing distance computation: ")

max_error = 0
for i in range(NO_TRY):
    
    coin1 = np.random.randint(0,5)
    coin2 = np.random.randint(0,5)
    
    obj1 = generate_rand_object(objtype = coin1 if coin1!=3 else 0)
    obj2 = generate_rand_object(objtype = coin2 if coin2!=3 else 0)
    
    _,_, d_cpp, _ = obj1.compute_dist(obj2,  p_init=[np.random.randn() for j in range(3)], tol=0.0001, no_iter_max=2000, mode='c++')
    _,_, d_py, _  = obj1.compute_dist(obj2,  p_init=[np.random.randn() for j in range(3)], tol=0.0001, no_iter_max=2000, mode='python')
    
    max_error = max(max_error,abs(d_cpp-d_py))
    
print("Max error distance computation: "+str(max_error))

print("-------------------------------------------------------------")
print("Testing check_free_config: ")

no_comp_failed = 0
for i in range(NO_TRY):
    
    obj = generate_rand_object()

    
    q_input = [np.random.randn() for j in range(7)]

    ok_cpp, _, _ = robot.check_free_config(q=q_input, obstacles=[obj], mode='c++')
    ok_py, _, _  = robot.check_free_config(q=q_input, obstacles=[obj], mode='python',  no_iter_max=5000)
    
    if not (ok_cpp==ok_py):
        no_comp_failed = no_comp_failed+1
        
print("Failed in "+str(no_comp_failed))




    


from utils import *
import numpy as np


def _dyn_model(self, q, qdot):
    n = len(self.links)

    # Error handling
    if not Utils.is_a_vector(qdot, n):
        raise Exception("The parameter 'qdot' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")
    # end error handling

    jj_geo, jac_geo, htm_com = self.jac_jac_geo(q=q, axis='com')

    i_mat_rot = []

    for i in range(n):
        i_mat = self.links[i].inertia_matrix + self.links[i].mass * Utils.S(self.links[i].com_coordinates) * Utils.S(
            self.links[i].com_coordinates)
        i_mat_rot.append(htm_com[i][0:3, 0:3] * i_mat * htm_com[i][0:3, 0:3].T)

    list_pjac_i_pk = []

    for i in range(n):
        list = []
        for k in range(n):
            pjac_i_pk = np.matrix(np.zeros((6, n)))
            for j in range(i + 1):
                pjac_i_pk[:, j] = jj_geo[i][j][:, k]

            list.append(pjac_i_pk)

        list_pjac_i_pk.append(list)

    list_pm_pk = []

    for k in range(n):
        pm_pk = np.zeros((n, n))
        for i in range(k, n):
            pm_pk += self.links[i].mass * list_pjac_i_pk[i][k][0:3, :].T * jac_geo[i][0:3, :]
            pm_pk += (list_pjac_i_pk[i][k][3:6, :].T + jac_geo[i][3:6, :].T * Utils.S(
                jac_geo[i][3:6, k])) * i_mat_rot[i] * jac_geo[i][3:6, :]

        list_pm_pk.append(pm_pk + (pm_pk).T)

    # Compute the M matrix

    dyn_m = np.zeros((n, n))

    for i in range(n):
        dyn_m += self.links[i].mass * jac_geo[i][0:3, :].T * jac_geo[i][0:3, :]
        dyn_m += jac_geo[i][3:6, :].T * i_mat_rot[i] * jac_geo[i][3:6, :]

    # Compute the C matrix

    dyn_c1 = np.matrix(np.zeros((n, 1)))
    dyn_c2 = np.matrix(np.zeros((n, 1)))
    qdot_v = np.matrix(qdot).reshape((n, 1))

    for j in range(n):
        dyn_c1 = dyn_c1 + (qdot_v[j,0] * list_pm_pk[j]) * qdot_v

    for i in range(n):
        dyn_c2[i] = 0.5 * qdot_v.T * list_pm_pk[i] * qdot_v

    dyn_c = (dyn_c1 - dyn_c2).reshape((n, 1))

    # Compute the G matrix

    dyn_g = np.zeros((n, 1))
    GRAVITY_ACC = 9.8
    for i in range(n):
        dyn_g += GRAVITY_ACC * self.links[i].mass * jac_geo[i][2, :].reshape((n, 1))

    return dyn_m, dyn_c, dyn_g



from utils import *
import numpy as np
from graphics.meshmaterial import *
from uaibot.simobjects.cylinder import *
from uaibot.simobjects.box import *

class Pedestrian(Cylinder):
    """
    Initialize a pedestrian agent.

    Parameters
    ----------
    ra : numpy array
        Position of the agent.
    g : numpy array
        Position of the agents goal.
    name : str, optional
        Name of the agent. The default is "pedestrian".
    color : str, optional
        Color of the agent. The default is "CornflowerBlue".
    height : float, optional
        Height of the agent. The default is 1.7.
    radius : float, optional
        Radius of the agent. The default is 0.3.
    va : numpy array, optional
        Velocity of the agent. The default is np.array([[0],[0]]).
    va0 : float, optional
        Desired velocity magnitude. The default is 1.3.
    ta : float, optional
        Relaxation time. The default is 1.
    ma : float, optional
        Agent mass. The default is 1.
    a_beta : float, optional
        Repulsive force amplitude relative to other pedestrians. The default is 5.
    b_beta : float, optional
        Repulsive force decay relative to other pedestrians. The default is 0.1.
    a_i : float, optional
        Repulsive force amplitude relative to obstacles. The default is 8.
    b_i : float, optional
        Repulsive force decay relative to obstacles. The default is 0.2.
    lambda_a : float, optional
        Directional sensitivity. The default is 1.
    """

    #######################################
    # Attributes
    #######################################

    @property
    def ra(self):
        return self._ra

    @ra.setter
    def ra(self, value):
        self._ra = value

    @property
    def g(self):
        return self._g

    @g.setter
    def g(self, value):
        self._g = value

    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, value):
        self._name = value

    @property
    def color(self):
        return self._color
    
    @color.setter
    def color(self, value):
        self._color = value
    
    @property
    def height(self):
        return self._height
    
    @height.setter
    def height(self, value):
        self._height = value

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = value

    @property
    def va(self):
        return self._va

    @va.setter
    def va(self, value):
        self._va = value

    @property
    def va0(self):
        return self._va0

    @va0.setter
    def va0(self, value):
        self._va0 = value

    @property
    def ta(self):
        return self._ta

    @ta.setter
    def ta(self, value):
        self._ta = value

    @property
    def ma(self):
        return self._ma

    @ma.setter
    def ma(self, value):
        self._ma = value

    @property
    def a_beta(self):
        return self._a_beta

    @a_beta.setter
    def a_beta(self, value):
        self._a_beta = value

    @property
    def b_beta(self):
        return self._b_beta

    @b_beta.setter
    def b_beta(self, value):
        self._b_beta = value

    @property
    def a_i(self):
        return self._a_i

    @a_i.setter
    def a_i(self, value):
        self._a_i = value

    @property
    def b_i(self):
        return self._b_i

    @b_i.setter
    def b_i(self, value):
        self._b_i = value

    @property
    def lambda_a(self):
        return self._lambda_a

    @lambda_a.setter
    def lambda_a(self, value):
        self._lambda_a = value

    @property
    def is_pedestrian(self):
        return True
    
    #######################################
    # Constructor
    #######################################

    def __init__(
        self,
        ra: np.array,
        g: np.array,
        name: str = "pedestrian",
        color: str = "CornflowerBlue",
        height: float = 1.7,
        radius: float = 0.3,
        va: np.array = np.array([[0], [0]]),
        va0: float = 1.3,
        ta: float = 1.0,
        ma: float = 1.0,
        a_beta: float = 5.0,
        b_beta: float = 0.4,
        a_i: float = 7.0,
        b_i: float = 0.3,
        lambda_a: float = 0.3,
    ):
        htm = np.eye(4)
        htm[0][3] = ra[0]
        htm[1][3] = ra[1]
        htm[2][3] = height / 2

        super().__init__(htm=htm, name=name, radius=radius, height=height, mass=1, color=color, opacity=1, mesh_material=None)
        self._ra = ra
        self._g = g
        self._name = name
        self._color = color
        self._height = height
        self._radius = radius
        self._va = va
        self._va0 = va0
        self._ta = ta
        self._ma = ma
        self._a_beta = a_beta
        self._b_beta = b_beta
        self._a_i = a_i
        self._b_i = b_i
        self._lambda_a = lambda_a

    def fag(self):
            """
            Compute the attractive force towards a goal.

            This function calculates the force required to steer an agent or object
            from its current velocity (va) toward a goal direction (ea0) at a
            desired velocity magnitude (va0). It uses a simple linear force model.
            """
            ea0 = (self._g - self._ra) / np.linalg.norm(self._ra - self._g)  # desired direction
            return self._ma * (1 / self._ta) * (self._va0 * ea0 - self._va)
    
    def fdaq(self, obstacle):
        """
        Compute the distance-dependent repulsive force.

        This function implements the exponential decay of repulsive force based on the
        distance vector d. The force magnitude decays exponentially with the norm of d,
        and its direction is along the unit vector of d.
        """
        d = obstacle.d(self._ra, self._radius)
        norm_d = np.linalg.norm(d)

        if obstacle.is_pedestrian:
            fdaq = self._a_beta * np.exp(-norm_d / self._b_beta) * (d / norm_d)
        else:
            fdaq = self._a_i * np.exp(-norm_d / self._b_i) * (d / norm_d)

        return fdaq
    
    def w(self, obstacle):
        """
        Compute the anisotropic weighting function for repulsive forces.

        This function introduces directional sensitivity to interactions
        based on the angle phi_aq, which is the angle between the agent's
        velocity vector and the direction of another entity. The parameter
        lambda_a modulates how much the agent is influenced by interactions
        from behind.
        """
        if self._va[0] == 0 and self._va[1] == 0:
            return 1 # if the agent is stationary, it is not influenced by interactions
        
        d = obstacle.d(self._ra, self._radius)
        phi_aq = np.arccos(
            np.dot(
                self._va.flatten() / np.linalg.norm(self._va),
                -d.flatten() / np.linalg.norm(d),
            )
        )

        return self._lambda_a + (1 - self._lambda_a) * (1 + np.cos(phi_aq)) / 2
    
    def d(self, rb, b_radius):
        """
        Compute the distance vector between two agents.
        This function calculates the distance vector between two agents
        based on their positions and radii.

        Parameters
        ----------
        rb : numpy array
            Position of the other agent.
        b_radius : float
            Radius of the other agent
        """

        d = (rb - self._ra) - (self._radius + b_radius) * (rb - self._ra) / np.linalg.norm(
            rb - self._ra
        )  # distance from obstacle to pedestrian
        return d
    

class ObstacleColumn(Cylinder):
    """
    Initialize an Obstacle Column.

    Parameters
    ----------
    ro : numpy array
        Position of the Obstacle.
    name : str, optional
        Name of the Obstacle. The default is "ObstacleColumn".
    color : str, optional
        Color of the Obstacle. The default is "DarkGrey".
    height : float, optional
        Height of the Obstacle. The default is 2.4.
    radius : float, optional
        Radius of the Obstacle. The default is 0.5.
    va : numpy array, optional
        Velocity of the Obstacle. The default is np.array([[0],[0]]).
    """

    #######################################
    # Attributes
    #######################################

    @property
    def ro(self):
        return self._ra

    @ro.setter
    def ro(self, value):
        self._ra = value

    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, value):
        self._name = value

    @property
    def color(self):
        return self._color
    
    @color.setter
    def color(self, value):
        self._color = value
    
    @property
    def height(self):
        return self._height
    
    @height.setter
    def height(self, value):
        self._height = value

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = value

    @property
    def va(self):
        return self._va

    @va.setter
    def va(self, value):
        self._va = value

    @property
    def is_pedestrian(self):
        return False
    
    #######################################
    # Constructor
    #######################################

    def __init__(
        self,
        ro: np.array,
        name: str = "ObstacleColumn",
        color: str = "DimGray",
        height: float = 2.4,
        radius: float = 0.5,
        va: np.array = np.array([[0], [0]])
    ):
        htm = np.eye(4)
        htm[0][3] = ro[0]
        htm[1][3] = ro[1]
        htm[2][3] = height / 2

        super().__init__(htm=htm, name=name, radius=radius, height=height, mass=1, color=color, opacity=1, mesh_material=None)
        self._ra = ro
        self._name = name
        self._color = color
        self._height = height
        self._radius = radius
        self._va = va

    
    def d(self, rb, b_radius):
        """
        Compute the distance vector between two agents.
        This function calculates the distance vector between two agents
        based on their positions and radii.

        Parameters
        ----------
        rb : numpy array
            Position of the other agent.
        b_radius : float
            Radius of the other agent
        """

        d = (rb - self._ra) - (self._radius + b_radius) * (rb - self._ra) / np.linalg.norm(
            rb - self._ra
        )  # distance from obstacle to pedestrian
        return d
    
class ObstacleThinWall(Box):
    """
    Initialize a thin wall obstacle.

    Parameters
    ----------
    start : numpy array
        Start position of the wall.
    end : numpy array
        End position of the wall.
    height : float, optional
        Height of the wall. The default is 2.4.
    """

    #######################################
    # Attributes
    #######################################

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        self._start = value

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        self._end = value

    @property
    def height(self):
        return self._height
    
    @height.setter
    def height(self, value):
        self._height = value

    @property
    def color(self):
        return self._color
    
    @color.setter
    def color(self, value):
        self._color = value

    @property
    def is_pedestrian(self):
        return False
    
    #######################################
    # Constructor
    #######################################

    def __init__(
        self,
        start: np.array,
        end: np.array,
        height: float = 2.4,
        color: str = "DimGray"
    ):
        self._start = start
        self._end = end
        self._height = height
        self._color = color

        wall_middle = (start + end) / 2
        wall_length = np.linalg.norm(end - start)
        wall_angle = np.arctan2(end[1][0] - start[1][0], end[0][0] - start[0][0])
        htm_trn = Utils.trn([wall_middle[0][0], wall_middle[1][0], height/2])
        htm_angle = Utils.rot([0, 0, 1], wall_angle)
        htm_total = htm_trn @ htm_angle
        super().__init__(htm=htm_total, name="", width=wall_length, height=height, depth=0.05, mass=1, color=color, opacity=1, mesh_material=None)
    
    #######################################
    # Methods
    #######################################

    def d(self, ra, ra_radius, retur_points=False):
        """
        computes (v, p0, c):
        - v is the smallest vector pointing from the segment ab to the circle
            with center o and radius r
        - p0 is the closest point on the segment ab to the circle center o
        - c is the corresponding closest point on the circle

        assumes the segment and circle do not intersect (d > r).

        parameters
        ----------
        ra : ndarray, shape (2,)
            position of the circle center
        ra_radius : float
            radius of the circle

        returns
        -------
        v : ndarray, shape (2,)
            smallest vector from the segment to the circle.
            if p0 is the point on the segment closest to o,
            then v = c - p0, where c is the circle point closest to p0.
        p0 : ndarray, shape (2,)
            closest point on the segment ab to the circle center o
        c : ndarray, shape (2,)
            closest point on the circle to the point p0
        """
        a = self.start.flatten()
        b = self.end.flatten()
        o = ra.flatten()

        ab = b - a
        ao = o - a
        norm_ab_sq = np.dot(ab, ab)

        # handle degenerate case: a == b
        if norm_ab_sq < 1e-15:
            # the segment is effectively a single point
            p0 = a
        else:
            # 1) parametric t_star for the infinite line that gives
            #    the closest approach of o onto line ab
            t_star = np.dot(ao, ab) / norm_ab_sq

            # 2) clamp t_star to [0, 1] to stay on segment ab
            t_0 = max(0.0, min(t_star, 1.0))

            # 3) closest point on the segment
            p0 = a + t_0 * ab

        # distance from p0 to o
        op0 = p0 - o
        d = np.linalg.norm(op0)

        # if truly no intersection, we expect d > r
        if d < ra_radius:
            # if intersection or numerical issues arise, handle gracefully
            return np.zeros(2), p0, p0

        # 4) closest point on the circle to p0 is on the ray o->p0
        c = o + (ra_radius / d) * op0

        # 5) smallest vector from p0 to c
        d = c - p0

        if retur_points:
            return d.reshape(2,1), p0.reshape(2,1), c.reshape(2,1)
        else:
            return d.reshape(2,1)


from typing import Union, List, TypeAlias, Tuple, TYPE_CHECKING
import numpy.typing as npt
import numpy as np
from typing import Optional

if TYPE_CHECKING:
    from uaibot.simobjects import (Ball, Box, ConvexPolytope, Cylinder, PointCloud, 
                                   Frame, RigidObject, Group, PointLight, Arrow, 
                                   HTMLDiv)
    from uaibot.robot import Robot
    from uaibot.graphics import MeshMaterial


Vector: TypeAlias = Union[List[float], Tuple[float], npt.NDArray[np.float64], np.matrix]
Matrix: TypeAlias = Union[npt.NDArray[np.float64], np.matrix]
HTMatrix: TypeAlias = np.matrix
MetricObject: TypeAlias = Union["Ball", "Box", "ConvexPolytope", "Cylinder", "PointCloud"]
GroupableObject: TypeAlias = Union["Ball", "Box", "Cylinder", "ConvexPolytope", "Frame",
                    "RigidObject", "Group", "Robot", "PointLight"]
SimObject: TypeAlias = Union["Ball", "Box", "Cylinder", "ConvexPolytope", "Frame",
                    "RigidObject", "Group", "Robot", "PointLight", "Arrow", "HTMLDiv"]





from robot.links import Link
import robot as rb
from utils import Utils
from graphics.meshmaterial import MeshMaterial
from graphics.model3d import Model3D
from simobjects.rigidobject import RigidObject
from simobjects.pointlight import PointLight
from simulation import Simulation
from simobjects.frame import Frame
import numpy as np

def map_color(color):
    if isinstance(color, list):
        colors = color
        if len(color) < 3:
            print("Jaco has 3 colors. Repeating the last one provided.")
            while len(colors) < 3:
                colors.append(color[-1])
    
    elif color is None:
        colors = ["#3e3f42", "#919090", "#1d1d1f"]
    
    else:
        colors = [color]*3

    return colors

def _create_jaco(htm=np.identity(4), name='', color=None, opacity=1, eef_frame_visible=True):
    colors = map_color(color)
    color1, color2, color3 = colors

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception(
            "The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")
    for color in colors:
        if (not Utils.is_a_color(color)):
            raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    pi = np.pi
    d1 = 0.2755
    d2 = 0.41
    d3 = 0.2073
    d4 = 0.0741
    d5 = 0.0741
    d6 = 0.16
    e2 = 0.0098
    aa = 30*pi/180
    sa = np.sin(aa)
    s2a = np.sin(2*aa)
    d4b = d3 + (sa/s2a) * d4
    d5b = (sa/s2a)*d4 + (sa/s2a)*d5
    d6b = (sa/s2a)*d5 + d6

    jaco_DH_theta = np.array([0,     0,    0,    0,    0,    0])
    jaco_DH_d = np.array([d1,    0,  -e2, -d4b, -d5b, -d6b])
    jaco_DH_a = np.array([0,    d2,    0,    0,    0,    0])
    jaco_DH_alpha = np.array([pi/2, pi, pi/2, 2*aa, 2*aa,    pi])

    jaco_DH_type = np.array([0, 0, 0, 0, 0, 0])
    link_info = np.array(
        [jaco_DH_theta, jaco_DH_d, jaco_DH_alpha, jaco_DH_a, jaco_DH_type])

    scale = 1
    n = link_info.shape[1]
    base_3d_obj = []
    mesh = MeshMaterial(metalness=0.8, clearcoat=0.9, roughness=0.3,
                        normal_scale=[0.5, 0.5], color=color1,
                        opacity=opacity, side="DoubleSide")
    mesh_ring = MeshMaterial(metalness=0, roughness=1, clearcoat=0, clearcoat_roughness=0.03, ior=1.45,
                             normal_scale=[0.5, 0.5], color=color2,
                             opacity=opacity, side="DoubleSide")
    mesh_nail = MeshMaterial(metalness=0, clearcoat=0, roughness=1,
                             normal_scale=[0.5, 0.5], color=color3,
                             opacity=opacity, side="DoubleSide")
   
    q0 = [pi, pi, pi, pi, 0, pi/2]

    Q00 = Utils.trn([0, 0, 0])
    Q01 = (Utils.rotz(link_info[0, 0] - pi/2 - q0[0]) * Utils.trn([0, 0, link_info[1, 0]])
           * Utils.rotx(link_info[2, 0]) * Utils.trn([link_info[3, 0], 0, 0]))
    Q02 = Q01 @ (Utils.rotz(link_info[0, 1] - pi/2 - np.deg2rad(25.002) + q0[1]) * Utils.trn(
        [0, 0, link_info[1, 1]]) * Utils.rotx(link_info[2, 1]) * Utils.trn([link_info[3, 1], 0, 0]))
    Q03 = Q02 @ (Utils.rotz(link_info[0, 2] - pi/2 + np.deg2rad(25.002) + q0[2]) * Utils.trn(
        [0, 0, link_info[1, 2]]) * Utils.rotx(link_info[2, 2]) * Utils.trn([link_info[3, 2], 0, 0]))
    Q04 = Q03 @ (Utils.rotz(link_info[0, 3] + pi/2 + q0[3]) * Utils.trn(
        [0, 0, link_info[1, 3]]) * Utils.rotx(link_info[2, 3]) * Utils.trn([link_info[3, 3], 0, 0]))
    Q05 = Q04 @ (Utils.rotz(link_info[0, 4] + pi + q0[4]) * Utils.trn(
        [0, 0, link_info[1, 4]]) * Utils.rotx(link_info[2, 4]) * Utils.trn([link_info[3, 4], 0, 0]))
    Q06 = Q05 @ Utils.rotz(link_info[0, 5] + pi/2 + q0[5]) * Utils.trn(
        [0, 0, link_info[1, 5]]) * Utils.rotx(link_info[2, 5]) * Utils.trn([link_info[3, 5], 0, 0])

    link0_mth = Utils.inv_htm(Q00)
    base_3d_obj = [
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/base.obj',
                scale=scale, htm=link0_mth, mesh_material=mesh),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/base_ring.obj',
                scale=scale, htm=link0_mth, mesh_material=mesh_ring)]
    
    link_3d_obj = []
    
    # Shoulder
    link1_mth = Utils.inv_htm(Q01)
    link_3d_obj.append([
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/shoulder.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/shoulder_ring.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh_ring),
    ])

    # Upper arm + elbow
    link2_mth = Utils.inv_htm(Q02)
    link_3d_obj.append([
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/upperarm.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/upperarm_ring.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh_ring),
    ])

    # Forearm
    link3_mth = Utils.inv_htm(Q03)
    link_3d_obj.append([
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/forearm.obj',
                scale=scale, htm=link3_mth, mesh_material=mesh),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/forearm_ring.obj',
                scale=scale, htm=link3_mth, mesh_material=mesh_ring),
    ])

    link4_mth = Utils.inv_htm(Q04)
    link_3d_obj.append([
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/wrist1.obj',
                scale=scale, htm=link4_mth, mesh_material=mesh),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/wrist1_ring.obj',
                scale=scale, htm=link4_mth, mesh_material=mesh_ring),
    ])

    link5_mth = Utils.inv_htm(Q05)
    link_3d_obj.append([
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/wrist2.obj',
                scale=scale, htm=link5_mth, mesh_material=mesh),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/wrist2_ring.obj',
                scale=scale, htm=link5_mth, mesh_material=mesh_ring),
    ])

    link6_mth = Utils.inv_htm(Q06)
    link_3d_obj.append([
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/gripper.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/handpalm.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/finger1_mounting.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/finger1_proximal.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/finger1_distal.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/finger1_nail.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_nail),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/finger2_mounting.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/finger2_proximal.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/finger2_distal.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/finger2_nail.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_nail),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/thumb_mounting.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/thumb_proximal.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/thumb_distal.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_ring),
        Model3D(url='https://raw.githubusercontent.com/fbartelt/robotics-experiments/main/models/jaco/thumb_nail.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh_nail),
    ])

    # com_coordinates = [[0.534615, 0, 0.15], [1.5353, 0, 0.15]]
    com_coordinates = [np.eye(4)[:-1, :] @ (Utils.inv_htm(Q01) @ np.array([-3.11506292e-3,  1.62075358e-5,  2.66810879e-1, 1]).reshape(-1, 1)),
                       np.eye(4)[:-1, :] @ (Utils.inv_htm(Q02) @ np.array(
                           [-0.00592762,  0.14709695,  0.5909634, 1]).reshape(-1, 1)),
                       np.eye(4)[:-1, :] @ (Utils.inv_htm(Q03) @ np.array(
                           [0.01162087, 0.04102689, 0.53732661, 1]).reshape(-1, 1)),
                       np.eye(4)[:-1, :] @ (Utils.inv_htm(Q04) @ np.array(
                           [0.00971901, -0.01057426,  0.44918022, 1]).reshape(-1, 1)),
                       np.eye(4)[:-1, :] @ (Utils.inv_htm(Q05) @ np.array(
                           [0.0097224, -0.03312561,  0.3785306, 1]).reshape(-1, 1)),
                       np.eye(4)[:-1, :] @ (Utils.inv_htm(Q06) @ np.array([0.0033009, -0.09643814,  0.32164165, 1]).reshape(-1, 1))]
    list_inertia_mat = []

    # Use parameters in MENDES, M. P. Computed torque-control of the Kinova JACO2 Arm. 2017
    list_mass = [0.767664, 0.99571300239, 0.79668391394,
                0.41737, 0.41737, 1.07541]
    I1 = np.array([[0.002231083822876,  0.000006815637176,  -0.000019787600863], 
                    [0,  0.000620733840723, -0.000306288710418], 
                    [0, 0,  0.002398007441187]])
    I1 = np.triu(I1, 1).T + I1
    list_inertia_mat.append(I1)
    I2 = np.array([[0.004162524943115,  -0.000000552736891,  -0.001487468585390], 
                    [0,  0.025495429281017, -0.000000277011102], 
                    [0, 0,  0.021736935450151]])
    I2 = np.triu(I2, 1).T + I2
    list_inertia_mat.append(I2)
    I3 = np.array([[0.002861254222538,  0.000000402773293,  0.000000566406981], 
                    [0,  0.002738656386387, -0.000344659168888], 
                    [0, 0,  0.000351242752544]])
    I3 = np.triu(I3, 1).T + I3
    list_inertia_mat.append(I3)
    I4 = np.array([[0.708476728234e-3,  0.008271643300e-3,  0.112833961462e-3], 
                    [0,  0.740496291632e-3, 0.000494273498e-3], 
                    [0, 0,  0.178193295188e-3]])
    I4 = np.triu(I4, 1).T + I4
    list_inertia_mat.append(I4)
    I5 = np.array([[0.827477604255e-3,  0.008281078147e-3,  -0.101690165377e-3], 
                    [0,  0.852081770282e-3, -0.000054222717e-3], 
                    [0, 0,  0.170777897817e-3]])
    I5 = np.triu(I5, 1).T + I5
    list_inertia_mat.append(I5)
    I6 = np.array([[0.004833755945304,  0.000004331179432,  0.000361596139440], 
                    [0,  0.004841503493061, -0.000002455131406], 
                    [0, 0,  0.000199821519482]])
    I6 = np.triu(I6, 1).T + I6
    list_inertia_mat.append(I6)
        
    links = []
    for i in range(n):
        links.append(
            Link(i,
                 theta=link_info[0, i],
                 d=link_info[1, i],
                 alpha=link_info[2, i],
                 a=link_info[3, i],
                 joint_type=link_info[4, i],
                 list_model_3d=link_3d_obj[i],
                #  com_coordinates=com_coordinates[i], # DEPRECATED PARAMETERS IN UAIBOT>=1.2.2
                #  mass=list_mass[i],
                #  inertia_matrix=list_inertia_mat[i],
            )
        )

    htm_n_eef = Utils.rotz(-pi) * Utils.rotx(0.3056*pi) * \
        Utils.rotx(0.3056*pi) * Utils.trn([0, 0, 0.052])
    htm_n_eef = Utils.trn([0, 0, 0])
    htm_base_0 = Utils.trn([0, 0, 0])

    # Create joint limits
    joint_limits = np.matrix(
        [
            [-3*np.pi, 3*np.pi],
            [-np.deg2rad(47), np.deg2rad(266)],
            [-np.deg2rad(19), np.deg2rad(322)],
            [-3*np.pi, 3*np.pi],
            [-3*np.pi, 3*np.pi],
            [-3*np.pi, 3*np.pi]
        ]
    )
    
    robot_ = rb.Robot(
        name,
        links,
        base_3d_obj,
        htm,
        htm_base_0,
        htm_n_eef,
        q0,
        eef_frame_visible=eef_frame_visible,
        joint_limits=joint_limits,
    )

    return robot_



from uaibot.simobjects.ball import *
from uaibot.simobjects.group import *
from uaibot.simobjects.cylinder import *
from uaibot.graphics.meshmaterial import *
from utils import *
import numpy as np
from typing import Optional, Tuple, List

GroupableObject: TypeAlias = Union["Ball", "Box", "Cylinder", "ConvexPolytope", "Frame",
                    "RigidObject", "Group", "Robot", "PointLight"]


class Frame:
    """
  A frame object.

  Parameters
  ----------
  htm : 4x4 numpy matrix
      The object's configuration.
      (default: the same as the current HTM).

  name : string
      The object's name.
      (default: '' (automatic)).

  size : positive float
      The axis sizes, in meters.
      (default: 0.3).

  axis_color : list of 3 HTML-compatible strings
      A list of 3 HTML-compatible strings, one for each axis.
      (default: ['red', 'lime', 'blue']).

  axis_names : list of 3 string
      The axis names.
      (default: ['x', 'y', 'z']).
  """

    #######################################
    # Attributes
    #######################################

    @property
    def size(self) -> float:
        """The axis size, in meters."""
        return self._size

    @property
    def name(self) -> str:
        """Name of the object."""
        return self._name

    @property
    def htm(self) -> "HTMatrix":
        """Object pose. A 4x4 homogeneous transformation matrix written is scenario coordinates."""
        return np.matrix(self._ball.htm)

    @property
    def axis_color(self) -> List[str]:
        """The axis colors. It is a list of 3 HTML-compatible colors."""
        return self._axis_color

    @property
    def axis_name(self) -> List[str]:
        """The axis names. It is a list of 3 strings."""
        return self._axis_name


    #######################################
    # Constructor
    #######################################

    def __init__(self, htm: "HTMatrix" =np.identity(4), name: str ="", size: float =0.3, 
                 axis_color: List[str] =['red', 'lime', 'blue'],
                 axis_names: List[str] =['x', 'y', 'z']) -> "Frame":

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(size) or size <= 0:
            raise Exception("The parameter 'size' should be a positive float.")

        if name=="":
            name="var_frame_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not (str(type(axis_color)) == "<class 'list'>") or (not (len(axis_color) == 3)):
            raise Exception("The parameter 'list' should be a a list of 3 HTML-compatible color strings.")

        for color in axis_color:
            if not Utils.is_a_color(color):
                raise Exception("The parameter 'list' should be a a list of 3 HTML-compatible color strings.")

        # end error handling

        self._name = name
        self._size = size
        self._axis_names = axis_names
        self._axis_color = axis_color
        self._ball = Ball(name="dummy_ball_" + name, htm=htm, radius=0.0001, mesh_material=MeshMaterial(opacity=0))
        self._max_time = 0
        
        cyl_x = Cylinder(radius=0.004, height=size, color = axis_color[0],htm=Utils.roty(np.pi/2)*Utils.trn([0,0,size/2]))
        cyl_y = Cylinder(radius=0.004, height=size, color = axis_color[1],htm=Utils.rotx(-np.pi/2)*Utils.trn([0,0,size/2]))
        cyl_z = Cylinder(radius=0.004, height=size, color = axis_color[2],htm=Utils.trn([0,0,size/2]))
        self._axis_group = Group([cyl_x, cyl_y, cyl_z])

        # Set initial total configuration
        self.set_ani_frame(np.matrix(htm))

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Axes with name '" + self.name + "': \n\n"
        string += " HTM: \n" + str(self.htm) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time: float, htm: Optional["HTMatrix"] = None) -> None:
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.
    htm : 4x4 numpy array
        The object's configuration.
        (default: the same as the current HTM).

    Returns
    -------
    None
    """

        self._ball.add_ani_frame(time, htm)
        self._axis_group.add_ani_frame(time, htm)
        self._max_time = self._ball._max_time

    # Set config. Restart animation queue
    def set_ani_frame(self, htm: Optional["HTMatrix"] = None) -> None:
        """
    Reset object's animation queue and add a single configuration to the
    object's animation queue.

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The object's configuration.
        (default: the same as the current HTM).

    Returns
    -------
    None
    """

        self._ball.set_ani_frame(htm)
        self._axis_group.set_ani_frame(htm)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""
        
        return self._axis_group.gen_code()

        # string = "\n"
        # string += "//BEGIN DECLARATION OF THE FRAME '" + self.name + "'\n\n"
        # string += self._ball.gen_code().replace("//USER INPUT GOES HERE", "")
        # string += "var var_axes_" + self.name + " = new AxesHelper(" + str(
        #     self.size) + ");\n"
        # string += "var_dummy_ball_" + self.name + ".shape.add(var_axes_" + self.name + ");\n"
        # string += "var_axes_" + self.name + ".setColors('" + \
        #           self.axis_color[0] + "', '" + self.axis_color[1] + "', '" + self.axis_color[2] + "');\n"
        # string += "sceneElements.push(var_dummy_ball_" + self.name + ");\n"
        # string += "//USER INPUT GOES HERE"
        # return string



import numpy as np
import os
from utils import *

import os
if os.environ['CPP_SO_FOUND']=="1":
    import uaibot_cpp_bind as ub_cpp

_INVHALFPI = 0.63660
_EPS = 1/10000000

def _vector_field_rn(q, curve, alpha, const_vel, is_closed, gamma, mode='auto'):

    n = np.shape(curve[0])[0]

    # Error handling
    if mode not in ['python','c++','auto']:
        raise Exception("The parameter 'mode' should be 'python,'c++', or 'auto'.")

    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_number(alpha) or alpha <= 0:
        raise Exception("The parameter 'alpha' should be a positive float.")

    if not Utils.is_a_number(const_vel):
        raise Exception("The parameter 'const_vel' should be a float.")

    if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
            raise Exception("c++ mode is set, but .so file was not loaded!")

    # end error handling

    if mode=='python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
        return _vector_field_rn_python(Utils.cvt(q), curve, alpha, const_vel, is_closed, gamma)
    else:
        vf_res =  ub_cpp.vectorfield_rn(Utils.cvt(q), curve, alpha, const_vel, is_closed, gamma)
        return Utils.cvt(vf_res.twist), vf_res.dist, vf_res.index



def _vector_field_rn_python(p, curve, alpha, const_vel, is_closed, gamma):


    vec_n, vec_t, min_dist, ind_min = _compute_ntd(curve, p)

    fun_g = _INVHALFPI * atan(alpha * np.sqrt(min_dist))
    fun_h = sqrt(max(1 - fun_g ** 2, 0))
    abs_const_vel = abs(const_vel)
    sgn = const_vel / (abs_const_vel + 0.00001)
    
    if is_closed:
        mult = 1
    else:
        per =  ind_min/len(curve)
        mult = min(gamma * (1.0-per),1.0)
    

    return abs_const_vel * (fun_g * vec_n + mult * sgn * fun_h * vec_t), min_dist, ind_min


def _compute_ntd(curve, p):
    min_dist = float('inf')
    ind_min = -1


    n = np.shape(curve[0])[0]

    pr = Utils.cvt(p)

    for i in range(len(curve)):
        dist_temp = np.linalg.norm(pr - Utils.cvt(curve[i]))
        if dist_temp < min_dist:
            min_dist = dist_temp
            ind_min = i



    vec_n = Utils.cvt(curve[ind_min]) - pr
    vec_n = vec_n / (np.linalg.norm(vec_n) + _EPS)

    if ind_min == len(curve) - 1:
        vec_t = Utils.cvt(curve[1]) - Utils.cvt(curve[ind_min])
    else:
        vec_t = Utils.cvt(curve[ind_min + 1]) - Utils.cvt(curve[ind_min])

    vec_t = vec_t / (np.linalg.norm(vec_t) + _EPS)
    vec_t = np.matrix(vec_t).reshape((n,1))

    return vec_n, vec_t, min_dist, ind_min



import uaibot as ub
import numpy as np
import matplotlib.pyplot as plt
import sys


def _control_demo_1():
    
    Tmax=20
    dt = 0.01
    K = 1.0
    v_speed = 0.3
    alpha = 1.0
    y_c = 0.5
    z_c = 0.6
    delta=0.3
    
    if ub.Utils.get_environment()=='Local':
        plt.figure()
        plt.xlim(-delta, delta)
        plt.ylim(z_c-delta, z_c+delta)
        plt.title("Right click to add points. Add at least 3. Left click to erase. Enter to finish")
        plt.xlabel("X")
        plt.ylabel("Z")
        plt.grid(True)

        points = plt.ginput(n=0, timeout=0)
        
        if len(points)<3:
            raise Exception("At least 3 points are necessary!")
            
        plt.close()  
    else:
        print("In Google Colab/Jupyter Notebook is not possible to draw your own curve... ")
        print("Run locally for this feature. Type at least 3 points you want in the format ")
        print("x,z ")
        print("in which "+str(round(-delta,2))+" <= x <= "+str(round(delta,2))+" and "+str(round(z_c-delta,2))+" <= z <= "+str(round(z_c+delta,2)))
        print("Write 'done' when you want to be over.")
        points = []
        while True:
            user_input = input("Enter point (x, z): ").strip()
            
            if user_input.lower() == 'done':
                break

            try:
                # Evaluate the tuple safely
                point = eval(user_input, {"__builtins__": None}, {})
                
                if isinstance(point, tuple) and len(point) == 2:
                    x, z = point
                    x = float(x)
                    z = float(z)
                    
                    if x>=-delta and x<=delta and z<=z_c+delta and z>=z_c-delta:
                        points.append((x, z))
                    else:
                        print("Error: "+str(round(-delta,2))+" <= x <= "+str(round(delta,2))+" and "+str(round(z_c-delta,2))+" <= z <= "+str(round(z_c+delta,2)))
                        
                else:
                    print("Invalid format. Please enter a tuple like (1.0, 2.0).")
            
            except Exception as e:
                print(f"Invalid input: {e}")


    s = [i / 1000 for i in range(1000)]

    target_curve = ub.Utils.interpolate(points, is_closed = True)(s)
    target_curve_mod = []
    for p in target_curve:
        target_curve_mod.append(np.matrix([p[0,0], y_c, p[1,0]]).T)



    robot = ub.Robot.create_kuka_lbr_iiwa()
    point_cloud = ub.PointCloud(points=target_curve_mod,size=0.02,color='yellow')

    sim = ub.Simulation.create_sim_grid([robot, point_cloud])
    sim.set_parameters(width=500,height=500,load_screen_color='#191919',background_color='#191919')
    
    rot_tg = robot.fkm()*ub.Utils.rotx(-np.pi/2)

    hist_p = []
    for i in range(round(Tmax/dt)):
        t = i*dt
        htm = robot.fkm(robot.q)
        p = htm[0:3,-1]
        v, dist, _ = ub.Robot.vector_field(p,target_curve_mod,alpha=alpha)
        r, jac_r = robot.task_function(q = robot.q, htm_tg = rot_tg)
        r[0:3,-1] = v_speed*v
        r[3:6,-1] = -K*r[3:6,-1]
        qdot = ub.Utils.dp_inv_solve(jac_r,r)
        
        robot.add_ani_frame(t,robot.q+qdot*dt)

        str_msg = "\rt = "+str(round(t,2))+" s, dist = "+str(round(1000*dist,2))+" mm"
        
        hist_p.append(p)

        sys.stdout.write(str_msg)
        sys.stdout.flush()
   
    plt.figure()
    plt.plot([p[0, 0] for p in target_curve], [p[1, 0] for p in target_curve])
    plt.scatter([p[0] for p in points], [p[1] for p in points], color='red')
    plt.plot([p[0, 0] for p in hist_p], [p[2, 0] for p in hist_p],color='green')

    dm = 0.2
    
    plt.xlim(-delta-dm, delta+dm)
    plt.ylim(z_c-delta-dm, z_c+delta+dm)
    plt.title("Tracked curve")
    plt.xlabel("X")
    plt.ylabel("Z")
    plt.grid(True)
    plt.show()
                
    return sim








from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_staubli_tx60(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    c = (np.pi / 180)
    link_info = [[0, 0, 0, 0, 0, 0], #"theta" rotation in z
                 [0, 0, 0.02, 0.31, 0, 0.07],  # "d" translation in z
                 [3.14 / 2, 0, 3.14 / 2, 3.14 / 2, -3.14 / 2, 0],  # "alfa" rotation in x
                 [0, 0.29, 0, 0, 0, 0],  # "a" translation in x 0.25
                 [0, 0, 0, 0, 0, 0]] #joint type

    n = 6
    scale = 1

    # Collision model
    col_model = [[], [], [], [], [], []]


    col_model[0].append(Cylinder(htm=Utils.trn([0, -0.15, 0]) * Utils.rotx(-3.14 / 2),
                                 name=name + "_C0_0", radius=0.1, height=0.45, color="red", opacity=0.3))
    col_model[1].append(Box(htm=Utils.trn([-0.15, 0, -0.125]) * Utils.rotz(3.14 / 2) * Utils.rotx(-3.14 / 2 - c * 7.3),
                            name=name + "_C1_0", width=0.18, depth=0.1, height=0.45, color="blue", opacity=0.3))
    col_model[2].append(Box(htm=Utils.trn([0, -0.02, 0]),
                            name=name + "_C2_0", width=0.17, depth=0.21, height=0.25, color="green", opacity=0.3))
    col_model[3].append(Box(htm=Utils.trn([0, -0.08, 0]),
                            name=name + "_C3_0", width=0.15, depth=0.27, height=0.13, color="orange", opacity=0.3))

    # Create 3d objects
    htm1 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, 1., -0.005], [0., -1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm2 = np.matrix([[0.0008, 0.1279, 0.9918, -0.31], [-1., 0.0001, 0.0008, 0.], [0., -0.9918, 0.1279, 0.02], [0., 0., 0., 1.]])
    htm3 = np.matrix([[-1., 0.0016, 0., -0.], [-0.0016, -1., 0., 0.005], [0., 0., 1., 0.], [0., 0., 0., 1.]])
    htm4 = np.matrix([[-1., -0.0016, 0., 0.], [0., 0.0008, 1., -0.31], [-0.0016, 1., -0.0008, 0.], [0., 0., 0., 1.]])

    base_3d_obj = [Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/base_link.stl',
        scale,
        np.identity(4),
        MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5],
                     color=color, opacity=opacity))]

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_1.stl',
            scale,
            htm1,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_2.stl',
            scale,
            htm2,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_3.stl',
            scale,
            htm3,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_4.stl',
            scale,
            htm4,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_5.stl',
            scale,
            np.identity(4),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#C0C0C0",
                         opacity=opacity))
         ]
    )
    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_6.stl',
            scale,
            np.identity(4),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="black",
                         opacity=opacity))
         ]
    )

    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration

    htm_base_0 = Utils.trn([0, 0, 0.375])

    q0 = [0, 3.14/2, 0, 0, 0, 0]

    #Create joint limits
    joint_limits = (np.pi/180)*np.matrix([[-180,180],[-127.5+90, 127.5+90],[-142.5+90, 142.5+90],[-270,270],[-121,132.5],[-270,270]])
    
    return base_3d_obj, links, htm_base_0, np.identity(4), q0, joint_limits



import os
import sys
import subprocess
import setuptools
import sysconfig
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install
from setuptools.command.develop import develop
from pathlib import Path

class CMakeExtension(Extension):
    """Defines a CMake extension for compiling C++ code."""
    def __init__(self, name, sourcedir=""):
        super().__init__(name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):
    """Runs CMake and Make to build the shared library."""
    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        python_include_dir = sysconfig.get_path("include")
        python_library_dir = sysconfig.get_config_var("LIBDIR")

        cmake_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}",
            f"-DPython3_EXECUTABLE={sys.executable}",
            f"-DPython3_INCLUDE_DIR={python_include_dir}",
            f"-DPython3_LIBRARY_DIR={python_library_dir}",
            "-DCMAKE_BUILD_TYPE=Release",
            "-DCMAKE_MAKE_PROGRAM=make",
            f"-DPython3_ROOT_DIR={os.path.dirname(os.path.dirname(sys.executable))}",
        ]


        build_temp = os.path.join(self.build_temp, ext.name)
        os.makedirs(build_temp, exist_ok=True)

        try:
            # Run CMake configuration
            subprocess.run(["cmake", ext.sourcedir] + cmake_args, 
                        cwd=build_temp, 
                        check=True,
                        capture_output=True,  # This captures the output
                        text=True)
            
            # Compile the C++ extension
            subprocess.run(["cmake", "--build", ".", "--config", "Release"],
                        cwd=build_temp,
                        check=True,
                        capture_output=True,  # This captures the output
                        text=True)
        
        except subprocess.CalledProcessError as e:
            print(f"CMake configuration failed with output:\n{e.stdout}\n{e.stderr}")
            raise
        # Explicitly copy the built library to the expected location
        lib_name = f"{ext.name}{sysconfig.get_config_var('EXT_SUFFIX')}"
        src_path = os.path.join(extdir, lib_name)
        dest_path = self.get_ext_fullpath(ext.name)
        os.makedirs(os.path.dirname(dest_path), exist_ok=True)
        if os.path.exists(src_path):
            self.copy_file(src_path, dest_path)

class CustomInstall(install):
    """Ensures the C++ extension is built during pip install."""
    def run(self):
        self.run_command("build_ext")
        install.run(self)

class CustomDevelop(develop):
    """Ensures the C++ extension is built during development install."""
    def run(self):
        self.run_command("build_ext")
        develop.run(self)

#Setup the installation
setup(
    ext_modules=[CMakeExtension("uaibot_cpp_bind", sourcedir="uaibot/c_implementation")],
    cmdclass={
        "build_ext": CMakeBuild,
        "install": CustomInstall,
        "develop": CustomDevelop,
    },
)



from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_abb_crb(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    link_info = [[       0,     0,      3.14,         0, 0, 0],  # "theta" rotation in z
                 [   0.265,     0,         0,     -0.47, 0, 0.101],  # "d" translation in z
                 [3.14 / 2,     0, -3.14 / 2, -3.14 / 2, -3.14 / 2, 0],  # "alfa" rotation in x
                 [       0, 0.444,      0.11,         0, 0.08, 0],  # "a" translation in x
                 [       0,      0,        0,         0, 0, 0]]  # joint type

    scale = 1
    n = 6

    # Collision model
    col_model = [[], [], [], [], [], []]

    col_model[0].append(Cylinder(htm= Utils.trn([0,0,0.025]),
                                 name=name + "_C0_0", radius=0.08, height=0.30, color="red", opacity=0.3))

    col_model[1].append(Cylinder(htm= Utils.trn([0,0,0.025]),
                                 name=name + "_C1_0", radius=0.08, height=0.30, color="blue", opacity=0.3))

    col_model[1].append(Box(htm= Utils.trn([-0.23,0,0.025+0.11]),
                                 name=name + "_C1_1", width=0.58, depth = 0.15, height=0.08, color="blue", opacity=0.3))

    col_model[2].append(Ball(htm= Utils.trn([-0.01,0,-0.065]),
                                 name=name + "_C2_0", radius=0.075, color="green", opacity=0.3))

    col_model[3].append(Cylinder(htm= Utils.trn([0,-0.3,0])*Utils.rotx(np.pi/2),
                                 name=name + "_C3_0", radius=0.065, height=0.20, color="yellow", opacity=0.3))

    col_model[3].append(Box(htm= Utils.trn([0,-0.1,-0.09]),
                                 name=name + "_C3_1", width=0.11, depth=0.32, height=0.08, color="yellow", opacity=0.3))

    col_model[3].append(Cylinder(htm= Utils.trn([0,0,-0.01]),
                                 name=name + "_C3_2", radius=0.065, height=0.20, color="yellow", opacity=0.3))

    col_model[4].append(Cylinder(htm= Utils.trn([0,0,-0.01]),
                                 name=name + "_C4_0", radius=0.065, height=0.24, color="magenta", opacity=0.3))
                          
    # Create 3d objects

    htm1 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, 1., 0.], [0., -1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm2 = np.matrix([[0.0008, 0., 1., -0.444], [-1., 0.0008, 0.0008, 0.], [-0.0008, -1., 0., 0.], [0., 0., 0., 1.]])
    htm3 = np.matrix([[0.0008, 0., 1., -0.11], [0., 1., 0., 0.], [-1., 0., 0.0008, 0.], [0., 0., 0., 1.]])
    htm4 = np.matrix([[0., -0.0008, -1., 0.], [1., 0.0008, 0., -0.47], [0.0008, -1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm5 = np.matrix([[0., -0.0016, 1., -0.08], [0.0008, -1., -0.0016, 0.], [1., 0.0008, 0., 0.], [0., 0., 0., 1.]])
    htm6 = np.matrix([[0.0008, 0., -1., 0.], [0., 1., 0., 0.], [1., 0., 0.0008, 0.], [0., 0., 0., 1.]])

    

    base_3d_obj = [
        Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/base_link.stl',
            scale,
            np.identity(4),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#606060",
                         opacity=opacity))]

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_1.stl',
                 scale,
                 htm1,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#606060",
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_2.stl',
                 scale,
                 htm2,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#606060",
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_3.stl',
                 scale,
                 htm3,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#606060",
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_4.stl',
                 scale,
                 htm4,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_5.stl',
                 scale,
                 htm5,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_6.stl',
                 scale,
                 htm6,
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="silver",
                              opacity=opacity))]
    )

    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    q0 = [0.000, 0.000, 3.14, 0.000, 0.000, 0.000]

    #Create joint limits
    joint_limits = (np.pi/180)*np.matrix([[-180,180],[-180,180],[-225-200+10,85-200-20],[-180,180],[-180,180],[-180,180]])


    return base_3d_obj, links, np.identity(4), np.identity(4), q0, joint_limits



from utils_test import *
import uaibot as ub
import time
import sys
import matplotlib.pyplot as plt
import matplotlib
import os

print("Testing free configuration")

robot = ub.Robot.create_magician_e6()
style = "top:" + str(0.9 * 720) + "px;right:" + str(0) + "px;width:" + str(
    960) + "px;position:absolute;text-align:center;color:white;background-color:#222224;font-smooth:always;font-family:arial"


htmldiv = ub.HTMLDiv(style = style)
projball1 = ub.Ball(radius=0.02,color='blue')
projball2 = ub.Ball(radius=0.02,color='red')

sim = ub.Simulation.create_sim_grid([robot, htmldiv, projball1, projball2])

Nobj = 10
Ntry = 500

for link in robot.links:
    for col_obj in link.col_objects:
        sim.add(col_obj[0])


all_obj = []
for i in range(Nobj):
    cont = True
    
    while cont:
        obj = generate_rand_object(color='red', size=0.2, spread=0.4)
        isfree, message, info = robot.check_free_config(q=robot.q, obstacles=[obj])
        cont = not isfree
    
    all_obj.append(obj)
        
    sim.add(obj)

sum_time = 0
max_time = 0
    
for i in range(Ntry):
    qmin = robot.joint_limit[:,0]
    qmax = robot.joint_limit[:,1]
    
    qnew = np.matrix(0*qmin)
    
    n = np.shape(qnew)[0]
    
    for j in range(n):
        qnew[j] = qmin[j]+np.random.uniform()*(qmax[j]-qmin[j])
        
    robot.add_ani_frame(i,qnew)
    robot.update_col_object(i)
    
    start = time.time() 
    isfree, message, info = robot.check_free_config(q=qnew, obstacles=all_obj)
    end =  time.time() 
    elapsed = end - start
    sum_time+=elapsed
    max_time=max(max_time,elapsed) 
    
    min_dist = 1e6
    for obj in all_obj:
        dr = robot.compute_dist(q=qnew, obj=obj).get_closest_item()
        if dr.distance < min_dist:
            min_dist = dr.distance
            point_obj = dr.point_object
            point_link = dr.point_link
        
    projball1.add_ani_frame(i,ub.Utils.trn(point_obj))
    projball2.add_ani_frame(i,ub.Utils.trn(point_link))
    
    htmldiv.add_ani_frame(i,message)
    
    sys.stdout.write("\rCompleted: "+str(round(100*i/Ntry))+"%")
    sys.stdout.flush()
    
    
    
print("\nAverage time took to check free configuration: "+str(round(1000*sum_time/Ntry,2))+" ms")
print("Max time took to check free configuration: "+str(round(1000*max_time,2))+" ms")

print("Test was a success!")   


current_folder = os.path.dirname(os.path.abspath(__file__))        
sim.save(current_folder,"test_free_config")
